# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

from configparser import ConfigParser

sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('.'))

if True:
    from changelog import generate_changelog

project = 'ESORM - ElasticSearch ORM'
# noinspection PyShadowingBuiltins
copyright = '2023, Adam Wallner'
author = 'Adam Wallner'

# Get version from setup.cfg
config = ConfigParser()
config.read('../setup.cfg')

# Generate changelog
changelog = generate_changelog('wallneradam/esorm')

version = config['metadata']['version']
release = config['metadata']['version']

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
    'sphinx_rtd_theme',
    'sphinx_rtd_dark_mode',
    'myst_parser',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_css_files = [
    'css/custom.css',
]
html_logo = '_static/img/esorm.svg'
html_favicon = '_static/img/favicon.ico'

source_suffix = ['.rst']

# user starts in light mode
default_dark_mode = False



"""
ESORM is an elasticsearch python ORM based on Pydantic
"""

from .error import (
    InvalidResponseError,
    InvalidModelError,
    NotFoundError,
)
from .model import (TModel, ESBaseModel, ESModel, ESModelTimestamp, Pagination, Sort, setup_mappings,
                    lazy_property, retry_on_conflict)
from .esorm import es, connect, get_es_version
from .fields import Field
from .bulk import ESBulk
from . import fields

__all__ = [
    "TModel", "ESBaseModel", "ESModel", "ESModelTimestamp", "ESBulk",
    'lazy_property', 'retry_on_conflict',
    "es",
    "NotFoundError",
    "InvalidModelError",
    "InvalidResponseError",
    "connect", "get_es_version",
    "setup_mappings",
    "Field",
    "fields",
    "error",
    "Pagination", "Sort",
]



from typing import List, Optional

import sys
import os
import asyncio
import subprocess
import importlib

import pytest
import logging

logger = logging.getLogger(__name__)
disable_loggers = ['elastic_transport.node_pool', 'elastic_transport.transport']


#
# Hooks
#

def pytest_configure():
    for logger_name in disable_loggers:
        logging.getLogger(logger_name).disabled = True


@pytest.fixture(scope="class")
def event_loop():
    """
    Use asyncio event loop in class scope
    """
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


#
# Fixtures
#

# noinspection PyProtectedMember
@pytest.fixture(scope="class")
async def esorm():
    """
    Simply import esorm module
    This way it can be removed too
    """
    esorm = importlib.import_module('esorm')
    # It should be empty for every class
    assert len(esorm.model._ESModelMeta.__models__) == 0
    yield esorm
    esorm.model._ESModelMeta.__models__.clear()
    for module_name in list(sys.modules):
        if module_name.startswith('esorm'):
            del sys.modules[module_name]


@pytest.fixture(scope="class")  # Test both ES 8.x and 7.x
async def docker_es(service):
    compose_file = os.path.join(os.path.dirname(__file__), 'docker-compose.yml')
    res = subprocess.run(['docker', 'compose', '-f', compose_file, 'up', '-d', service], check=True)
    logger.info(f"Started service: {service}")
    assert res.returncode == 0
    yield
    subprocess.run(['docker', 'compose', '-f', compose_file, 'down', service, '-v'], check=True)
    logger.info(f"Stopped service: {service}")


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
async def es(docker_es, esorm, service):
    """
    ElasticSearch fixture for version 8.x
    """
    es = await esorm.connect(hosts=["http://localhost:9200"], wait=True)

    es_version = await esorm.get_es_version()
    major, _, _ = map(int, es_version.split('.'))
    if service == 'es7x':
        assert major == 7
    elif service == 'es8x':
        assert major == 8

    major, minor, _ = map(int, es_version.split('.'))
    assert es is not None
    yield es
    await es.close()


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_python(esorm):
    """
    Model to test python types
    """
    from typing import Literal
    from datetime import datetime, date, time
    from pydantic import UUID4, HttpUrl, FilePath

    class PythonFieldModel(esorm.ESModel):
        f_str: str
        f_int: int
        f_float: float
        f_bool: bool
        f_datetime: datetime
        f_date: date
        f_time: time
        f_literal: Literal['a', 'b', 'c']

        # Pydantic types which may annotated to python types
        f_uuid4: UUID4
        f_file_path: FilePath
        f_http_url: HttpUrl

    return PythonFieldModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_es(esorm):
    """
    Model to test ES types
    """

    class ESFieldModel(esorm.ESModel):
        f_keyword: esorm.fields.keyword
        f_text: esorm.fields.text
        f_binary: esorm.fields.binary
        f_byte: esorm.fields.byte
        f_short: esorm.fields.short
        f_int32: esorm.fields.int32
        f_long: esorm.fields.long
        f_unsigned_long: esorm.fields.unsigned_long
        f_float16: esorm.fields.float16
        f_float32: esorm.fields.float32
        f_double: esorm.fields.double
        f_geo_point: esorm.fields.geo_point

    return ESFieldModel


@pytest.fixture(scope="class")
def model_es_optional(esorm):
    """
    Model to test optional ES types
    """
    from pydantic import PositiveInt

    class ESOptionalFieldModel(esorm.ESModel):
        f_keyword: Optional[esorm.fields.keyword] = None
        f_text: Optional[esorm.fields.text] = None
        f_binary: Optional[esorm.fields.binary] = esorm.Field(None, index=False)
        f_byte: Optional[esorm.fields.byte] = None
        f_short: Optional[esorm.fields.short] = None
        f_int32: Optional[esorm.fields.int32] = None
        f_long: Optional[esorm.fields.long] = None
        f_unsigned_long: Optional[esorm.fields.unsigned_long] = None
        f_float16: Optional[esorm.fields.float16] = None
        f_float32: Optional[esorm.fields.float32] = None
        f_double: Optional[esorm.fields.double] = None
        f_geo_point: Optional[esorm.fields.geo_point] = None

        f_positive_int: Optional[PositiveInt] = None

    return ESOptionalFieldModel


@pytest.fixture(scope="class")
def model_es_optional_new_syntax(esorm):
    """
    Model to test optional ES types with new syntax
    """

    if sys.version_info >= (3, 10):
        class ESOptionalFieldNewSyntaxModel(esorm.ESModel):
            f_keyword: esorm.fields.keyword | None = None
            f_text: esorm.fields.text | None = None
            f_binary: esorm.fields.binary | None = esorm.Field(None, index=False)
            f_byte: esorm.fields.byte | None = None
            f_short: esorm.fields.short | None = None
            f_int32: esorm.fields.int32 | None = None
            f_long: esorm.fields.long | None = None
            f_unsigned_long: esorm.fields.unsigned_long | None = None
            f_float16: esorm.fields.float16 | None = None
            f_float32: esorm.fields.float32 | None = None
            f_double: esorm.fields.double | None = None
            f_geo_point: esorm.fields.geo_point | None = None

        return ESOptionalFieldNewSyntaxModel

    return None


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_timestamp(esorm):
    """
    Model to test timestamp
    """

    class TimestampModel(esorm.ESModelTimestamp):
        f_str: str
        f_int: int = 1

    return TimestampModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_config(esorm):
    """
    Model to test config
    """

    class ConfigModel(esorm.ESModel):
        class ESConfig:
            index = 'custom_index'
            id_field = 'custom_id'
            default_sort = [{'f_str': {'order': 'desc'}}]
            settings = dict(
                number_of_shards=6,
                number_of_replicas=1,
                refresh_interval="5s",
            )

        custom_id: str
        """ The id of the document """

        f_str: str
        """ Field to test """

        @property
        def __routing__(self) -> str:
            return self.custom_id + '_routing'

    return ConfigModel


@pytest.fixture(scope="class")
def model_index_template(esorm):
    """
    Model to test index template
    """

    class IndexTemplateModel(esorm.ESModel):
        class ESConfig:
            index = 'custom_index_template'

        f_str: str

    return IndexTemplateModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_with_id(esorm):
    """
    Model to test id field
    """

    class IdModel(esorm.ESModel):
        id: str
        f_str: str

    return IdModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_with_int_id(esorm):
    """
    Model to test int id field
    """

    class IntIdModel(esorm.ESModel):
        class ESConfig:
            id_field = 'custom_id'

        custom_id: int
        f_str: str

    return IntIdModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_with_prop_id(esorm):
    """
    Model to test property id field
    """

    class PropIdModel(esorm.ESModel):
        @property
        def __id__(self) -> int:
            return self.custom_id + 1000

        custom_id: int
        f_str: str

    return PropIdModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_nested(esorm, model_timestamp):
    """
    Model to test nested fields
    """

    class NestedFieldModel(esorm.ESModel):
        f_nested: model_timestamp
        f_float: float = 0.5

    return NestedFieldModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
def model_lazy_prop(esorm):
    """
    Model to test lazy properties
    """

    class LazyPropModel(esorm.ESModel):
        class ESConfig:
            lazy_property_max_recursion_depth = 1

        f_str: str

        # noinspection PyUnresolvedReferences
        @esorm.lazy_property
        async def same_f_strs(self) -> List['LazyPropModel']:
            return await self.search_by_fields({'f_str': self.f_str})

    return LazyPropModel


# noinspection PyUnresolvedReferences
@pytest.fixture(scope="class")
async def model_nested_binary(esorm):
    """
    Model to test nested binary fields
    """

    class BinaryModel(esorm.ESBaseModel):
        f_binary: Optional[esorm.fields.binary] = esorm.Field(None, index=False)

    class NestedBinaryModel(esorm.ESModel):
        f_nested: BinaryModel

    await esorm.setup_mappings()

    return BinaryModel, NestedBinaryModel


@pytest.fixture(scope="class")
async def model_nested_base_model(esorm):
    """
    Model to test nested base model
    """
    from asyncio import sleep

    class NestedBaseModel(esorm.ESBaseModel):
        f_str: str
        f_int: int

        @esorm.lazy_property
        async def lazy_prop(self) -> str:
            await sleep(0.1)
            return f'lazy(f_str={self.f_str}, f_int={self.f_int})'

    class NestedBaseModelModel(esorm.ESModel):
        f_nested: NestedBaseModel

    await esorm.setup_mappings()

    return NestedBaseModel, NestedBaseModelModel


@pytest.fixture(scope="class")
async def model_base_model_parent(esorm):
    """
    Model to test base model as a parent class
    """

    class BaseModelParent(esorm.ESBaseModel):
        class ESConfig:
            id_field = 'f_str'

        f_str: str
        f_int: int

    class BaseModelParentModel(BaseModelParent, esorm.ESModel):
        f_float: float

    await esorm.setup_mappings()

    return BaseModelParentModel



"""
ElasticSearch ORM main module
"""
from typing import Optional, cast, Union, List, Mapping

import asyncio

from elasticsearch import AsyncElasticsearch
from elastic_transport import NodeConfig

from .logger import logger


class _ESProxy:
    """
    ElasticSearch client proxy
    """
    __client__: Optional[AsyncElasticsearch] = None

    def set_client(self, client):
        self.__client__ = client

    def __getattr__(self, name):
        if self.__client__ is None:
            if name == '__wrapped__':  # This is needed for pdoc3
                return object.__getattribute__(self, name)
            raise ValueError(f"ElasticSearch client has not been set yet, please call connect! ({name})")
        return getattr(self.__client__, name)


# Global client proxy
es = cast(AsyncElasticsearch, _ESProxy())

__all__ = ['es', 'connect', 'get_es_version']


async def connect(hosts: Union[str, List[Union[str, Mapping[str, Union[str, int]], NodeConfig]]],
                  *args, wait=False, **kwargs) -> Optional[AsyncElasticsearch]:
    """
    Connect to ElasticSearch

    :param hosts: ElasticSearch hosts to connect, either a list a mapping, or a single string
    :param args: Other AsyncElasticsearch arguments
    :param wait: Wait for AsyncElasticsearch to be ready
    :param kwargs: Other AsyncElasticsearch keyword arguments
    :return: AsyncElasticsearch client instance
    """
    cast(_ESProxy, es).set_client(AsyncElasticsearch(hosts=hosts, *args, **kwargs))

    try:
        if wait:
            # Wait for ES to start, this will block until ES is ready to prevent routers start early
            while True:
                status = await es.ping()
                if not status:
                    logger.info("Waiting for ElasticSearch to be ready…")
                    await asyncio.sleep(2.0)
                else:
                    break

        return es

    except asyncio.CancelledError:
        return None


async def get_es_version() -> str:
    """
    Get ElasticSearch version

    :return: ElasticSearch version
    """
    info = await es.info()
    return info['version']['number']



"""
Elasticsearch query type definitions for ESORM
"""
from typing import TypedDict, List, Dict, Union
from typing_extensions import TypeAlias
from . import aggs


class ESRange(TypedDict, total=False):
    """
    Range query structure
    """
    gt: Union[int, float, str]
    """ Greater than """
    gte: Union[int, float, str]
    """ Greater than or equal """
    lt: Union[int, float, str]
    """ Less than """
    lte: Union[int, float, str]
    """ Less than or equal """


class FieldRange(TypedDict):
    """
    Range qyery field
    """
    range: Dict[str, ESRange]
    """ Range query structure """


class ESTerm(TypedDict, total=False):
    """
    Represents the parameters for a term query in Elasticsearch.
    """
    value: Union[str, int, float]
    """ The value to search for. """
    boost: Union[int, float]
    """ Optional boosting value for the query. """


class FieldTerm(TypedDict):
    """
    Represents a term query for exact value matching in Elasticsearch.
    """
    term: Dict[str, ESTerm]
    """ Term query structure """


class FieldTerms(TypedDict):
    """
    Represents a terms query for exact value matching in Elasticsearch.
    """
    terms: Dict[str, List[Union[str, int, float]]]
    """ Terms query structure """


class ESMatch(TypedDict, total=False):
    """
    Represents the parameters for a match query in Elasticsearch.
    """
    query: Union[str, int, float]
    """ The value to search for. """
    operator: str
    """ The operator to use for the query. """
    boost: Union[int, float]
    """ Optional boosting value for the query. """
    analyzer: str
    """ Optional analyzer to use for the query. """
    fuzziness: Union[int, str]
    """ Optional fuzziness value for the query. """
    prefix_length: int
    """ Optional prefix length for the query. """
    max_expansions: int
    """ Optional maximum number of expansions for the query. """
    zero_terms_query: str
    """ Optional zero terms query for the query. """


class FieldMatch(TypedDict):
    """
    Represents a match query for matching based on the provided text in Elasticsearch.
    """
    match: Dict[str, ESMatch]
    """ Match query structure """


class ESMatchPhrase(TypedDict, total=False):
    """
    Represents the parameters for a match_phrase query in Elasticsearch.
    """
    query: str
    """ The value to search for. """
    analyzer: str
    """ Optional analyzer to use for the query. """
    boost: Union[int, float]
    """ Optional boosting value for the query. """
    slop: int
    """ Optional slop value for the query. """


class FieldMatchPhrase(TypedDict):
    """
    Represents a match_phrase query for exact phrase matching in Elasticsearch.
    """
    match_phrase: Dict[str, ESMatchPhrase]
    """ Match phrase query structure """


class ESExists(TypedDict):
    """
    Represents an exists query to check if a field exists.
    """
    field: str
    """ The field to check. """


class FieldExists(TypedDict):
    """
    Represents an exists query to check if a field exists in Elasticsearch.
    """
    exists: ESExists
    """ Exists query structure """


class ESWildcard(TypedDict, total=False):
    """
    Represents a wildcard query for pattern matching in Elasticsearch.
    """
    value: str
    """ The pattern to search for. e.g., "te?t" or "test*" """
    boost: float
    """ Optional boosting value for the query """
    rewrite: str
    """ Optional, method used to rewrite the query (e.g., "constant_score", "scoring_boolean") """
    case_insensitive: bool
    """ Optional, whether the query is case insensitive. """


class FieldWildcard(TypedDict):
    """
    Represents a wildcard query for pattern matching in Elasticsearch.
    """
    wildcard: Dict[str, ESWildcard]
    """ Wildcard query structure """


class ESPrefix(TypedDict, total=False):
    """
    Represents a prefix query for prefix matching in Elasticsearch.
    """
    value: str
    """ The prefix to search for. """
    boost: float
    """ Optional boosting value for the query """
    rewrite: str
    """ Optional, method used to rewrite the query (e.g., "constant_score", "scoring_boolean") """


class FieldPrefix(TypedDict):
    """
    Represents a prefix query for prefix matching in Elasticsearch.
    """
    prefix: Dict[str, ESPrefix]
    """ Prefix query structure """


class ESFuzzy(TypedDict, total=False):
    """
    Represents a fuzzy query for approximate matching in Elasticsearch.
    """
    value: str
    """ The value to search for. """
    fuzziness: Union[int, str]
    """ Fuzziness value for the query """
    prefix_length: int
    """ Prefix length for the query """
    max_expansions: int
    """ Maximum number of expansions for the query """
    transpositions: bool
    """ Whether to allow transpositions for the query """
    boost: float
    """ Optional boosting value for the query """


class FieldFuzzy(TypedDict):
    """
    Represents a fuzzy query for approximate matching in Elasticsearch.
    """
    fuzzy: Dict[str, ESFuzzy]
    """ Fuzzy query structure """


class ESGeoDistance(TypedDict, total=False):
    """
    Represents a geo_distance query for distance-based geospatial queries in Elasticsearch.
    """
    distance: Union[str, float]
    """ The distance to search for. """
    distance_type: str
    """ The distance type to use for the query. """
    validation_method: str
    """ The validation method to use for the query. """
    location_field: str
    """ The field containing the location to search from. """
    location: Union[
        Dict[str, float],
        str
    ]
    """ The location to search from. """


class FieldGeoDistance(TypedDict):
    """
    Represents a geo_distance query for distance-based geospatial queries in Elasticsearch.
    """
    geo_distance: Dict[str, ESGeoDistance]
    """ Geo distance query structure """


class ESMatchAll(TypedDict, total=False):
    """
    Represents a match_all query for matching all documents in Elasticsearch.
    """
    boost: float
    """ Optional boosting value for the query """


class FieldMatchAll(TypedDict):
    """
    Represents a match_all query for matching all documents in Elasticsearch.
    """
    match_all: ESMatchAll
    """ Match all query structure """


class ESMatchNone(TypedDict):
    """
    Represents a match_none query for matching no documents in Elasticsearch.
    """
    pass


class FieldESMatchNone(TypedDict):
    """
    Represents a match_none query for matching no documents in Elasticsearch.
    """
    match_none: ESMatchNone
    """ Match none query structure """


#: Represents must queries in Elasticsearch
ESMust: TypeAlias = List[
    Union[
        FieldRange, FieldTerm, FieldTerms, FieldMatch, FieldMatchPhrase, FieldExists, FieldWildcard, FieldPrefix,
        FieldFuzzy, FieldGeoDistance,
        FieldMatchAll, FieldESMatchNone,
        'FieldBool'
    ]
]
""" Represents must queries in Elasticsearch """

#: Represents filter queries in Elasticsearch
ESFilter: TypeAlias = List[
    Union[
        FieldRange, FieldTerm, FieldTerms, FieldMatch, FieldMatchPhrase, FieldExists, FieldWildcard, FieldPrefix,
        FieldFuzzy, FieldGeoDistance,
        FieldMatchAll, FieldESMatchNone,
        'FieldBool'
    ]
]
""" Represents filter queries in Elasticsearch """

#: Represents should queries in Elasticsearch
ESShould: TypeAlias = List[
    Union[
        FieldRange, FieldTerm, FieldTerms, FieldMatch, FieldMatchPhrase, FieldExists, FieldWildcard, FieldPrefix,
        FieldFuzzy, FieldGeoDistance,
        FieldMatchAll, FieldESMatchNone,
        'FieldBool'
    ]
]
""" Represents should queries in Elasticsearch """

#: Represents must_not queries in Elasticsearch
ESMustNot: TypeAlias = List[
    Union[
        FieldRange, FieldTerm, FieldTerms, FieldMatch, FieldMatchPhrase, FieldExists, FieldWildcard, FieldPrefix,
        FieldFuzzy, FieldGeoDistance,
        FieldMatchAll, FieldESMatchNone,
        'FieldBool'
    ]
]
""" Represents must_not queries in Elasticsearch """


class ESBool(TypedDict, total=False):  # We use total=False, because not every key is required
    """
    Bool query structure
    """
    must: ESMust
    """ Must queries """
    filter: ESFilter
    """ Filter queries """
    should: ESShould
    """ Should queries """
    must_not: ESMustNot
    """ Must not queries """
    minimum_should_match: Union[int, str]
    """ Minimum number of should queries to match """
    boost: float
    """ Boosting value for the query """


class FieldBool(TypedDict):
    """
    Represents a bool query for combining other queries in Elasticsearch.
    """
    bool: ESBool
    """ Bool query structure """


class ESQuery(TypedDict, total=False):
    """
    Elasticsearch query structure
    """
    bool: ESBool
    """ Bool query structure """
    match: Dict[str, ESMatch]
    """ Match query structure """
    match_phrase: Dict[str, ESMatchPhrase]
    """ Match phrase query structure """
    term: Dict[str, ESTerm]
    """ Term query structure """
    prefix: Dict[str, ESPrefix]
    """ Prefix query structure """
    fuzzy: Dict[str, ESFuzzy]
    """ Fuzzy query structure """
    wildcard: Dict[str, ESWildcard]
    """ Wildcard query structure """
    geo_distance: Dict[str, ESGeoDistance]
    """ Geo distance query structure """
    exists: ESExists
    """ Exists query structure """
    match_all: ESMatchAll
    """ Match all query structure """
    match_none: ESMatchNone
    """ Match none query structure """
    aggs: aggs.ESAggs
    """ Aggregations query structure """



"""
ElasticSearch aggregation type definitions for ESORM
"""
from typing import TypedDict, Dict, List, Union
from typing_extensions import TypeAlias


#
# Queries
#

class ESAggTermParams(TypedDict, total=False):
    """
    Represents the parameters for a terms aggregation in Elasticsearch.
    """
    field: str
    """ The field to aggregate on. """
    size: int
    """ The number of buckets to return. """
    order: Dict[str, str]
    """ The order of the buckets. """


class ESAggFieldParams(TypedDict):
    """
    Represents field parameter in Elasticsearch.
    """
    field: str
    """ The field to aggregate on. """


class ESAggExtendedBounds(TypedDict):
    """
    Represents the parameters for extended bounds in Elasticsearch.
    """
    min: int
    """ The minimum value. """
    max: int
    """ The maximum value. """


class ESAggHistogramParams(TypedDict, total=False):
    """
    Represents the parameters for a histogram aggregation in Elasticsearch.
    """
    field: str
    """ The field to aggregate on. """
    interval: int
    """ The interval of the histogram. """
    min_doc_count: int
    """ The minimum number of documents in a bucket. """
    extended_bounds: ESAggExtendedBounds
    """ The extended bounds of the histogram. """


class ESAgg(TypedDict, total=False):
    """
    Holds all types of aggregations supported
    """
    terms: ESAggTermParams
    """ Terms aggregation """
    avg: ESAggFieldParams
    """ Average aggregation """
    sum: ESAggFieldParams
    """ Sum aggregation """
    min: ESAggFieldParams
    """ Minimum aggregation """
    max: ESAggFieldParams
    """ Maximum aggregation """
    histogram: ESAggHistogramParams
    """ Histogram aggregation """


ESAggs: TypeAlias = Dict[str, ESAgg]
""" ElasticSearch aggregations type definition """


#
# Responses
#

class ESAggBucketResponse(TypedDict):
    """
    Represents a single bucket in a bucket aggregation.
    """
    key: str
    """ The key of the bucket. """
    doc_count: int
    """ The number of documents in this bucket. """


class ESAggTermsResponse(TypedDict):
    """
    Represents the response for a terms aggregation.
    """
    buckets: List[ESAggBucketResponse]
    """ A list of buckets in the terms aggregation. """


class ESAggValueResponse(TypedDict):
    """
    Represents the response for an average, min, or max aggregation.
    """
    value: float
    """ The average, min, or max value. """


class ESAggHistogramBucketresponse(TypedDict):
    """
    Represents a bucket in a histogram aggregation.
    """
    key: float
    """ Numeric key corresponding to the bucket's range. """
    doc_count: int
    """ The number of documents in this bucket."""


class ESAggHistogramResponse(TypedDict):
    """
    Represents the response for a histogram aggregation.
    """
    buckets: List[ESAggHistogramBucketresponse]
    """ A list of buckets in the histogram aggregation. """


ESAggsResponse: TypeAlias = Dict[str, Union[
    ESAggValueResponse,
    ESAggTermsResponse,
    ESAggHistogramResponse
]]
""" ElasticSearch aggregations response type definition """



"""
This module contains the ESModel classes and related functions
"""
from typing import (TypeVar, Any, Dict, Optional, Tuple, Type, Union, get_args, get_origin, List, Callable,
                    Awaitable, Literal)
from typing_extensions import TypedDict, Annotated

import asyncio
import ast
import inspect
import textwrap
import traceback
from contextvars import ContextVar
from datetime import datetime, date, time
from functools import wraps

import elasticsearch
from elasticsearch import NotFoundError as ElasticNotFoundError, ConflictError as ElasticConflictError
from pydantic import main as pydantic_main
from pydantic import BaseModel, ConfigDict
from pydantic.fields import Field, PrivateAttr
# noinspection PyProtectedMember
from pydantic.fields import FieldInfo  # It is just not in __all__ of pydantic.fields, but we strongly need it
from pydantic_core import Url

from uuid import UUID
from pathlib import Path

from .utils import snake_case, utcnow
from .aggs import ESAggs, ESAggsResponse

from .error import InvalidResponseError, NotFoundError
from .esorm import es, get_es_version
from .query import ESQuery
from .response import ESResponse

from .logger import logger

__all__ = [
    'TModel',
    'ESBaseModel',
    'ESModel',
    'ESModelTimestamp',
    'Pagination', 'Sort',
    'setup_mappings',
    'create_index_template',
    'set_default_index_prefix',
    'set_max_lazy_property_concurrency',
    'lazy_property',
    'retry_on_conflict'
]

#
# Global variables and types
#

# noinspection PyProtectedMember
_model_construction = getattr(pydantic_main, '_model_construction')
ModelMetaclass = _model_construction.ModelMetaclass

_default_index_prefix = 'esorm_'

# Map python types to ES type
_pydantic_type_map = {
    str: 'keyword',  # Str is defaulted to keyword
    int: 'long',
    float: 'double',
    bool: 'boolean',
    datetime: 'date',
    date: 'date',
    time: 'date',  # Time is stored as datetime, 1970-01-01 + time
    # Other python types
    UUID: 'keyword',
    Path: 'keyword',
    # Pydantic core types
    Url: 'keyword',
}

# TModel type variable
TModel = TypeVar('TModel', bound='ESModel')

# Context variable to store the current recursion depth of lazy properties
_lazy_property_recursion_depth: ContextVar[int] = ContextVar('_lazy_property_recursion_depth', default=0)
_lazy_semaphore_concurrency = 16
_lazy_property_semaphore: ContextVar[asyncio.Semaphore] = ContextVar('_lazy_property_semaphore',
                                                                     default=asyncio.Semaphore(
                                                                         _lazy_semaphore_concurrency))


#
# Monkey patching to support field descriptions from docstrings
#

def _description_from_docstring(model: Type[BaseModel]):
    """
    Set undefined field descriptions from variable docstrings

    :param model: The model to set the descriptions
    """
    try:
        source = textwrap.dedent(inspect.getsource(model))
        module = ast.parse(source)
        assert isinstance(module, ast.Module)
        class_def = module.body[0]
        assert isinstance(class_def, ast.ClassDef)
        if len(class_def.body) < 2:
            return
    except OSError:
        return

    for last, node in zip(class_def.body, class_def.body[1:]):
        try:
            if not (isinstance(last, ast.AnnAssign) and isinstance(last.target, ast.Name) and
                    isinstance(node, ast.Expr)):
                continue

            info = model.model_fields[last.target.id]
            if info.description is not None:
                continue

            doc_node = node.value
            if isinstance(doc_node, ast.Constant):  # 'regular' variable doc string
                docstring = doc_node.value.strip()
            else:
                raise NotImplementedError(doc_node)

            info.description = docstring

        except KeyError:
            pass


def _patch_set_model_fields():
    """
    Monkey patchon _model_construction.set_model_fields to set undefined field descriptions from docstrings
    """
    orig_set_model_fields = _model_construction.set_model_fields

    def set_model_fields(model: Type[BaseModel], bases: Tuple[Type[Any], ...], config_wrapper: Any,
                         types_namespace: Dict[str, Any]) -> None:
        orig_set_model_fields(model, bases, config_wrapper, types_namespace)
        _description_from_docstring(model)

    _model_construction.set_model_fields = set_model_fields


_patch_set_model_fields()


#
# ElasticSearch models
#

def set_default_index_prefix(default_index_prefix: str):
    """
    Set default index prefix we use for model and index creation

    :param default_index_prefix: The default index prefix
    """
    global _default_index_prefix
    _default_index_prefix = default_index_prefix


class _ESModelMeta(ModelMetaclass):
    """
    ESModel Metaclass
    """

    # All model classes collected
    __models__: Dict[str, Type['ESModel']] = {}

    # noinspection PyUnresolvedReferences
    def __new__(cls: Type[ModelMetaclass], name: str, bases: Tuple[type, ...],
                namespace: Dict[str, Any], **kwds: Any):
        model: Type[BaseModel] = super().__new__(cls, name, bases, namespace, **kwds)
        if name not in ("ESModel", "ESModelTimestamp", "ESBaseModel"):
            is_model = issubclass(model, ESModel)

            # ESConfig inheritance
            m_dict = {k: v for k, v in ESModel.ESConfig.__dict__.items() if k[0] != '_'}
            if bases and 'ESConfig' in bases[0].__dict__:
                m_dict.update({k: v for k, v in bases[0].ESConfig.__dict__.items() if k[0] != '_'})
            del m_dict['index']  # It is only allowed to be set on the actual model class
            if 'ESConfig' in model.__dict__:
                m_dict.update({k: v for k, v in model.ESConfig.__dict__.items() if k[0] != '_'})
            m_dict['_lazy_properties'] = {}

            # Create (new) ESConfig class inside the class
            model.ESConfig = type('ESConfig', (object,), dict(m_dict))

            # Set default index name if not already set
            if is_model and not getattr(model.ESConfig, 'index', None):
                # Default index is the name of the class in snake_case
                model.ESConfig.index = _default_index_prefix + snake_case(name)

            # If there is an 'id' field, set it as id_field
            if is_model and 'id' in model.model_fields.keys():
                model.ESConfig.id_field = 'id'

            # Add to models
            if is_model:
                cls.__models__[model.ESConfig.index] = model

            # Collect lazy properties
            for attr_name, attr in namespace.items():
                # Support computed fields
                if attr.__class__.__name__ == 'PydanticDescriptorProxy':
                    attr = getattr(attr, 'wrapped')
                # Is it a lazy property?
                if isinstance(attr, property) and hasattr(attr.fget, '__lazy_property__'):
                    # noinspection PyProtectedMember
                    model.ESConfig._lazy_properties[attr_name] = getattr(attr.fget, '__lazy_property__')

        return model


class ESBaseModel(BaseModel, metaclass=_ESModelMeta):
    """
    Base class for Elastic

    It is useful for nested models, if you don't need the model in ES mappings
    """

    class ESConfig:
        """
        ESBaseModel Config

        This is just for lazy properties, to make ESBasemodel compatible with them
        """

        lazy_property_max_recursion_depth: int = 1
        """ Maximum recursion depth of lazy properties """

        _lazy_properties: Dict[str, Callable[[], Awaitable[Any]]] = {}
        """ Lazy property async function definitions """

    model_config = ConfigDict(
        str_strip_whitespace=True,
        extra="forbid",
        populate_by_name=True,
        arbitrary_types_allowed=True,
        ser_json_bytes='base64',
        validate_assignment=True,
    )

    async def calc_lazy_properties(self):
        """
        (re)Calculate lazy properties
        """
        _lazy_semaphore = _lazy_property_semaphore.get()
        # noinspection PyProtectedMember
        for attr_name, attr in self.ESConfig._lazy_properties.items():
            async with _lazy_semaphore:
                # If we use create_task, this creates a new context, so changed contextvars live only upward
                res = await asyncio.create_task(attr(self))
                setattr(self, '_' + attr_name, res)

        # Calc lazy properties for nested models
        for k, v in self.__dict__.items():
            if isinstance(v, ESBaseModel):
                await v.calc_lazy_properties()


class ESModel(ESBaseModel):
    """
    ElasticSearch Base Model
    """

    _id: Optional[str] = PrivateAttr(None)
    """ The ES id of the document (it is always a string) """

    _routing: Optional[str] = PrivateAttr(None)
    """ The routing of the document """

    _version: Optional[int] = PrivateAttr(None)
    """ The version of the document """

    _primary_term: Optional[int] = PrivateAttr(None)
    """ The primary term of the document """

    _seq_no: Optional[int] = PrivateAttr(None)
    """ The sequence number of the document """

    class ESConfig:
        """ ESModel Config """
        index: Optional[str] = None
        """ The index name """

        id_field: Optional[str] = None
        """ The name of the 'id' field """

        default_sort: Optional[List[Dict[str, Dict[str, str]]]] = None
        """ Default sort """

        settings: Optional[Dict[str, Any]] = None
        """ Index settings """

        lazy_property_max_recursion_depth: int = 1
        """ Maximum recursion depth of lazy properties """

        _lazy_properties: Dict[str, Callable[[], Awaitable[Any]]] = {}
        """ Lazy property async function definitions """

    @property
    def __id__(self) -> str:
        """
        The id of the document

        This can be overridden to make computed ids

        :return: The id of the document
        """
        return getattr(self, self.ESConfig.id_field or '_id')

    @property
    def __routing__(self) -> Optional[str]:
        """
        Shard route name

        :return: Shard route name
        """
        return None

    @classmethod
    async def call(cls: Type[TModel], method_name, *, wait_for=None, index: Optional[str] = None, **kwargs) -> dict:
        """
        Call an elasticsearch method

        This is a low level ES method call, it is not recommended to use this directly.

        :param method_name: The name of the method to call
        :param wait_for: Waits for all shards to sync before returning response
        :param index: The index name, if not set, it will use the index from ESConfig
        :param kwargs: The arguments to pass to the method
        :return: The result dictionary from ElasticSearch
        """
        kwargs = dict(kwargs)
        method = getattr(es, method_name)
        index = index or cls.ESConfig.index
        if wait_for is not None:
            kwargs['refresh'] = "wait_for"
        if 'request_timeout' not in kwargs:
            kwargs['request_timeout'] = 60

        return await method(index=index, **kwargs)

    @classmethod
    def _recursive_convert_to_es(cls: Type[TModel], data: dict, _level=0):
        """ Recursively modify data for Elasticsearch """
        for k, v in data.items():
            # Encode datetime fields
            if isinstance(v, datetime):
                # Update ESTimestamp fields
                if _level != 0 and k == 'modified_at':
                    v = utcnow()
                elif _level != 0 and k == 'created_at' and v is None:
                    v = utcnow()
                data[k] = v.replace(tzinfo=None).isoformat() + 'Z'

            # Convert date fields
            elif isinstance(v, date):
                data[k] = v.isoformat()

            # Convert time fields
            elif isinstance(v, time):
                data[k] = datetime.combine(datetime(1970, 1, 1), v).isoformat()

            # Convert UUID, Path, Url fields
            elif isinstance(v, UUID) or isinstance(v, Path) or isinstance(v, Url):
                data[k] = str(v)

            # Convert subclasses
            elif isinstance(v, dict):
                cls._recursive_convert_to_es(v, _level + 1)

    @classmethod
    def _recursive_exclude(cls: Type[TModel], m: BaseModel) -> Dict[str, Union[bool, dict]]:
        """ Recursively exclude computed fields """
        _exclude: Dict[str, Union[bool, dict]] = {k: True for k in m.model_computed_fields.keys()}
        for k, v in m:
            if k in _exclude:
                continue
            if isinstance(v, BaseModel):
                res = cls._recursive_exclude(v)
                if res:
                    _exclude[k] = res
        return _exclude

    def to_es(self, **kwargs) -> dict:
        """
        Generates a dictionary equivalent to what ElasticSearch returns in the '_source' property of a response.

        It automatically removes the id field from the document if it is set in ESConfig.id_field to prevent
        duplication of the id field.

        :param kwargs: Pydantic's model_dump parameters
        :return: The dictionary for ElasticSearch
        """
        kwargs = dict(kwargs)

        # Update exclude field with computed fields
        exclude = kwargs.get('exclude', {})
        exclude.update(self._recursive_exclude(self))
        kwargs['exclude'] = exclude
        # Dump model to dict
        d = self.model_dump(**kwargs)

        self._recursive_convert_to_es(d)
        return d

    @classmethod
    def _recursive_convert_from_es(cls: Type[TModel], data: dict):
        """ Recursively convert data from Elasticsearch """
        for k, v in data.items():
            try:
                field = cls.model_fields[k]
                # Convert time fields
                if field.annotation == time and isinstance(v, str):
                    data[k] = datetime.fromisoformat(v).time()

                # Dict fields
                elif isinstance(v, dict):
                    cls._recursive_convert_from_es(v)
            except KeyError:
                pass

    def update_from_es(self, data: Dict[str, Any]):
        """
        Update the model from ElasticSearch data

        :param data: The ElasticSearch data
        :raises esorm.error.InvalidResponseError: Returned when _id or _source is missing from data
        """
        if not data:
            return None

        source: Optional[dict] = data.get("_source", None)
        # Get id field
        _id = data.get("_id", None)
        if not source or not _id:
            raise InvalidResponseError

        for k, v in source.items():
            if k in self.__fields_set__:
                setattr(self, k, v)

        self._recursive_convert_from_es(source)

        # Set routing field
        _routing = data.get("_routing", None)
        setattr(self, '_routing', _routing)

        # Set version field
        _version = data.get("_version", None)
        setattr(self, '_version', _version)
        # Set primary term field
        _primary_term = data.get("_primary_term", None)
        setattr(self, '_primary_term', _primary_term)
        # Set seq_no field
        _seq_no = data.get("_seq_no", None)
        setattr(self, '_seq_no', _seq_no)

    @classmethod
    def from_es(cls: Type[TModel], data: Dict[str, Any]) -> Optional[TModel]:
        """
        Returns an ESModel from an elasticsearch document that has _id, _source

        :param data: Elasticsearch document that has _id, _source
        :raises esorm.error.InvalidResponseError: Returned when _id or _source is missing from data
        :return: The ESModel instance
        """
        if not data:
            return None

        source: Optional[dict] = data.get("_source", None)
        # Get id field
        _id = data.get("_id", None)
        if not source or not _id:
            raise InvalidResponseError

        # Add id field to document
        if source is not None and cls.ESConfig.id_field:
            source[cls.ESConfig.id_field] = _id

        # Concert to Pydantic model
        cls._recursive_convert_from_es(source)
        obj = cls(**source)

        # Set id field
        setattr(obj, '_id', _id)

        # Set routing field
        _routing = data.get("_routing", None)
        setattr(obj, '_routing', _routing)

        # Set version field
        _version = data.get("_version", None)
        setattr(obj, '_version', _version)
        # Set primary term field
        _primary_term = data.get("_primary_term", None)
        setattr(obj, '_primary_term', _primary_term)
        # Set seq_no field
        _seq_no = data.get("_seq_no", None)
        setattr(obj, '_seq_no', _seq_no)

        return obj

    async def save(self, *, wait_for=False, pipeline: Optional[str] = None, routing: Optional[str] = None) -> str:
        """
        Save document into elasticsearch.

        If document already exists, existing document will be updated as per native elasticsearch index operation.
        If model has id (Config.id_field or __id__), this will be used as the elasticsearch _id. The id field will be
        removed from the document before indexing.
        If no id is provided, then document will be indexed and elasticsearch will generate a suitable id that will be
        populated on the returned model.

        :param wait_for: Waits for all shards to sync before returning response - useful when writing
                         tests. Defaults to False.
        :param pipeline: Pipeline to use for indexing
        :param routing: Shard routing value
        :return: The new document's ID, it is always a string, even if the id field is an integer
        """
        kwargs = dict(
            document=self.to_es(),
            wait_for=wait_for,
        )

        kwargs['id'] = self.__id__
        if self.ESConfig.id_field:
            del kwargs['document'][self.ESConfig.id_field]

        if pipeline is not None:
            kwargs['pipeline'] = pipeline

        kwargs['routing'] = routing if routing is not None else self.__routing__

        if self._primary_term is not None:
            kwargs['if_primary_term'] = self._primary_term
        if self._seq_no is not None:
            kwargs['if_seq_no'] = self._seq_no

        es_res = await self.call('index', **kwargs)

        # Update private fields
        self._id = es_res.get('_id', None)
        self._version = es_res.get('_version', None)
        self._primary_term = es_res.get('_primary_term', None)
        self._seq_no = es_res.get('_seq_no', None)
        # Return the new document's ID
        return self._id

    # noinspection PyShadowingBuiltins
    @classmethod
    async def get(cls: Type[TModel], id: Union[str, int, float], *, routing: Optional[str] = None,
                  index: Optional[str] = None) -> TModel:
        """
        Fetches document and returns ESModel instance populated with properties.

        :param id: Document id
        :param routing: Shard routing value
        :param index: Index name, if not set, it will use the index from ESConfig3663
        :raises esorm.error.NotFoundError: Returned if document not found
        :return: ESModel object
        """
        try:
            es_res = await cls.call('get', routing=routing, id=id, index=index)
            return await _lazy_process_results(cls.from_es(es_res))
        except ElasticNotFoundError:
            raise NotFoundError(f"Document with id {id} not found")

    async def delete(self, *, wait_for=False, routing: Optional[str] = None):
        """
        Deletes document from ElasticSearch.

        :param wait_for: Waits for all shards to sync before returning response - useful when writing
                         tests. Defaults to False.
        :param routing: Shard routing value
        :raises esorm.error.NotFoundError: Returned if document not found
        :raises ValueError: Returned when id attribute missing from instance
        """
        kwargs = dict(id=self.__id__)
        if self._primary_term is not None:
            kwargs['if_primary_term'] = self._primary_term
        if self._seq_no is not None:
            kwargs['if_seq_no'] = self._seq_no
        try:
            await self.call('delete', wait_for=wait_for,
                            routing=routing if routing is not None else self.__routing__,
                            **kwargs)
        except ElasticNotFoundError:
            raise NotFoundError(f"Document with id {self.__id__} not found!")

    async def reload(self, *, routing: Optional[str] = None) -> TModel:
        """
        Reloads the document from ElasticSearch

        :param routing: Shard routing value
        :raises esorm.error.NotFoundError: Returned if document not found
        """
        kwargs = dict(id=self.__id__, routing=routing if routing is not None else self.__routing__)
        try:
            es_res = await self.call('get', **kwargs)
            self.update_from_es(es_res)
            return await _lazy_process_results(self)
        except ElasticNotFoundError:
            raise NotFoundError(f"Document with id {id} not found")

    @classmethod
    async def _search(cls: Type[TModel],
                      query: Optional[ESQuery] = None,
                      *,
                      page_size: Optional[int] = None,
                      page: Optional[int] = None,
                      sort: Optional[Union[list, str]] = None,
                      routing: Optional[str] = None,
                      aggs: Optional[ESAggs] = None,
                      index: Optional[str] = None,
                      **kwargs) -> ESResponse:
        """
        Raw ES search method

        :param query: ElasticSearch query dict
        :param page_size: Pagination page size
        :param page: Pagination page num, 1st page is 1
        :param sort: Name of field to be sorted, or sort term list of dict, if not specified, model's default sort will
                     be used, or no sorting
        :param routing: Shard routing value
        :param aggs: Aggregations
        :param index: Index name, if not set, it will use the index from ESConfig
        :param kwargs: Other search API params
        :return: Raw ES response.
        """
        if isinstance(sort, str):
            sort = [{sort: {'order': 'asc'}}]
        elif sort is None and cls.ESConfig.default_sort is not None:
            sort = cls.ESConfig.default_sort

        if page_size is not None and page is None:
            page = 1

        return await cls.call('search', query=query,
                              from_=((page - 1) * page_size) if page_size is not None else 0,
                              size=page_size, sort=sort, routing=routing,
                              aggs=aggs,
                              index=index,
                              seq_no_primary_term=True, version=True,
                              **kwargs)

    @classmethod
    async def search(cls: Type[TModel], query: ESQuery, *,
                     page_size: Optional[int] = None,
                     page: Optional[int] = None,
                     sort: Optional[Union[list, str]] = None,
                     routing: Optional[str] = None,
                     res_dict: bool = False,
                     index: Optional[str] = None,
                     **kwargs) -> Union[List[TModel], Dict[str, TModel]]:
        """
        Search Model with query dict

        :param query: ElasticSearch query dict
        :param page_size: Pagination page size
        :param page: Pagination page num, 1st page is 1
        :param sort: Name of field to be sorted, or sort term list of dict, if not specified, model's default sort will
                     be used, or no sorting
        :param routing: Shard routing value
        :param res_dict: If the result should be a dict with id as key and model as value instead of a list of models
        :param index: Index name, if not set, it will use the index from ESConfig
        :param kwargs: Other search API params
        :return: The result list
        """
        res = await cls._search(query, page_size=page_size, page=page, sort=sort, routing=routing,
                                index=index, **kwargs)
        try:
            if res_dict:
                res = {hit['_id']: cls.from_es(hit) for hit in res['hits']['hits']}
            else:
                res = [cls.from_es(hit) for hit in res['hits']['hits']]
            return await _lazy_process_results(res)
        except KeyError:
            return []

    @classmethod
    async def search_one(cls: Type[TModel], query: ESQuery, *, routing: Optional[str] = None,
                         index: Optional[str] = None, **kwargs) -> Optional[TModel]:
        """
        Search Model and return the first result

        :param query: ElasticSearch query dict
        :param routing: Shard routing value
        :param index: Index name, if not set, it will use the index from ESConfig
        :param kwargs: Other search API params
        :return: The first result or None if no result
        """
        res = await cls.search(query, page_size=1, routing=routing, index=index, **kwargs)
        if len(res) > 0:
            return res[0]
        else:
            return None

    @staticmethod
    def create_query_from_dict(fields: Dict[str, Union[str, int, float]]) -> ESQuery:
        """
        Creates a query dict from a dictionary of fields and values

        :param fields: A dictionary of fields and values to search by
        :return: A query dict
        """
        return {
            'bool': {
                'must': [{
                    'match': {
                        k: {'query': v, 'operator': 'and'},
                    }
                } for k, v in fields.items()]
            }
        }

    @classmethod
    async def search_by_fields(cls: Type[TModel],
                               fields: Dict[str, Union[str, int, float]],
                               *,
                               page_size: Optional[int] = None,
                               page: Optional[int] = None,
                               sort: Optional[Union[list, str]] = None,
                               routing: Optional[str] = None,
                               aggs: Optional[ESAggs] = None,
                               res_dict: bool = False,
                               index: Optional[str] = None,
                               **kwargs) -> List[TModel]:
        """
        Search Model by fields as key-value pairs

        :param fields: A dictionary of fields and values to search by
        :param page_size: Pagination page size
        :param page: Pagination page num, 1st page is 1
        :param sort: Name of field to be sorted, or sort term list of dict, if not specified,
                     model's default sort will be used, or no sorting
        :param routing: Shard routing value
        :param aggs: Aggregations
        :param res_dict: If the result should be a dict with id as key and model as value instead of a list of models
        :param index: Index name, if not set, it will use the index from ESConfig
        :param kwargs: Other search API params
        :return: The result list
        """
        query = cls.create_query_from_dict(fields)
        return await cls.search(query, page_size=page_size, page=page, sort=sort, routing=routing,
                                aggs=aggs, res_dict=res_dict, index=index, **kwargs)

    @classmethod
    async def search_one_by_fields(cls: Type[TModel],
                                   fields: Dict[str, Union[str, int, float]],
                                   *, routing: Optional[str] = None,
                                   aggs: Optional[ESAggs] = None,
                                   index: Optional[str] = None,
                                   **kwargs) -> Optional[TModel]:
        """
        Search Model by fields as key-value pairs and return the first result

        :param fields: A dictionary of fields and values to search by
        :param routing: Shard routing value
        :param aggs: Aggregations
        :param index: Index name, if not set, it will use the index from ESConfig
        :param kwargs: Other search API params
        :return: The first result or None if no result
        """
        query = cls.create_query_from_dict(fields)
        return await cls.search_one(query, routing=routing, aggs=aggs, index=index, **kwargs)

    @classmethod
    async def all(cls: Type[TModel], index: Optional[str] = None, **kwargs) -> List[TModel]:
        """
        Get all documents

        :param kwargs: Other search API params
        :param index: Index name, if not set, it will use the index from ESConfig
        :return: The result list
        """
        return await cls.search({'match_all': {}}, index=index, **kwargs)

    @classmethod
    async def aggregate(cls: Type[TModel],
                        aggs: ESAggs,
                        *,
                        query: Optional[ESQuery] = None,
                        routing: Optional[str] = None,
                        index: Optional[str] = None,
                        **kwargs) -> ESAggsResponse:
        """
        Aggregate Model with aggregation dict
        Before aggregation the model can be filtered by query dict.

        :param aggs: Aggregation dict
        :param query: ElasticSearch query dict
        :param routing: Shard routing value
        :param index: Index name, if not set, it will use the index from ESConfig
        :param kwargs: Other search API params
        :return: The result list
        """
        try:
            res = await cls._search(query, aggs=aggs, routing=routing, page_size=0, index=index, **kwargs)
            return res['aggregations']
        except KeyError:
            return {}


class ESModelTimestamp(ESModel):
    """
    Model which stores `created_at` and `modified_at` fields automatcally.
    """
    created_at: Optional[datetime] = Field(None, description="Creation date and time")
    modified_at: Optional[datetime] = Field(default_factory=utcnow, description="Modification date and time")

    async def save(self, *, wait_for=False, force_new=False, pipeline: Optional[str] = None,
                   routing: Optional[str] = None) -> str:
        """
        Save document into elasticsearch.

        If document already exists, existing document will be updated as per native elasticsearch index operation.
        If model has id (Meta.id_field or __id__), this will be used as the elasticsearch _id. The id field will be
        removed from the document before indexing.
        If no id is provided, then document will be indexed and elasticsearch will generate a suitable id that will be
        populated on the returned model.

        :param wait_for: Waits for all shards to sync before returning response - useful when writing
            tests. Defaults to False.
        :param force_new: It is assumed to be a new document, so created_at will be set to current time
                          (it is no more necessary, because created_at is set to current time if it is None.
                          It is here for backward compatibility)
        :param pipeline: Pipeline to use for indexing
        :param routing: Shard routing value
        :return: The new document's ID
        """
        self.modified_at = utcnow()
        # Set created_at if not already set
        if force_new or not self.created_at:
            self.created_at = self.modified_at
        return await super().save(wait_for=wait_for, pipeline=pipeline, routing=routing)


#
# Lazy properties
#

def set_max_lazy_property_concurrency(concurrency: int):
    """
    Set the maximum concurrency of processing lazy properties

    If this is not set, the default is 16.

    :param concurrency: The maximum concurrency
    """
    global _lazy_semaphore_concurrency
    _lazy_semaphore_concurrency = concurrency
    _lazy_property_semaphore.set(asyncio.Semaphore(concurrency))


async def _lazy_process_results(res: Union[List[ESModel], ESModel, Dict[str, ESModel]]) \
        -> Union[List[ESModel], ESModel, Dict[str, ESModel]]:
    """
    Process the results of ES query to calculate lazy properties recursively

    :param res: The result of the endpoint
    :return: The result of the endpoint
    """
    tasks = []

    if isinstance(res, ESBaseModel):
        tasks.append(asyncio.create_task(res.calc_lazy_properties()))

    elif isinstance(res, list):
        for r in res:
            tasks.append(asyncio.create_task(r.calc_lazy_properties()))

    elif isinstance(res, dict):
        for r in res.values():
            tasks.append(asyncio.create_task(r.calc_lazy_properties()))

    else:
        raise TypeError(f"Invalid return type: {type(res)}")

    # Wait for tasks to be ready
    await asyncio.gather(*tasks)

    return res


def lazy_property(func: Callable[[], Awaitable[Any]]):
    """
    Decorator for lazy properties

    Lazy properties computed after search from ES

    :param func: The async function to decorate
    :return: The decorated function
    """
    assert inspect.iscoroutinefunction(func), \
        f"The function {func.__name__} must be a coroutine function"
    assert inspect.signature(func).return_annotation is not inspect.Signature.empty, \
        f"The function {func.__name__} must have a return annotation"

    @wraps(func)
    def wrapper(self):
        # Initialize call stack if not exists
        if not hasattr(self, '__lazy_call_stack__'):
            self.__lazy_call_stack__ = []

        # Return the property with underscore prefix
        try:
            return getattr(self, '_' + func.__name__)
        except AttributeError:
            return None

    @wraps(func)
    async def async_wrapper(self, *args, **kwargs):
        # Check recursion depth
        depth = _lazy_property_recursion_depth.get()
        if depth >= self.ESConfig.lazy_property_max_recursion_depth:
            logger.warning(f"Recursion depth exceeded for {self.__class__.__name__}.{func.__name__}")
            return None
        # Increase recursion depth
        _lazy_property_recursion_depth.set(depth + 1)
        _lazy_property_semaphore.set(asyncio.Semaphore(_lazy_semaphore_concurrency))
        # Call the original function
        return await func(self, *args, **kwargs)

    # Create a property from it
    prop = property(wrapper)
    # Set the original function as __lazy_property_func__
    setattr(wrapper, '__lazy_property__', async_wrapper)

    return prop


#
# Optimistic concurrency control
#

def retry_on_conflict(max_retries=-1):
    """
    Decorator for optimistic concurrency control

    :param max_retries: The maximum number of retries, -1 for infinite
    :return: The decorated function
    """

    def decorator(func: Callable):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            retries = 0
            while True:
                try:
                    return await func(*args, **kwargs)
                except ElasticConflictError:
                    if max_retries == -1 or retries < max_retries:
                        retries += 1
                        logger.warning(f"Optimistic concurrency control conflict, retrying {retries}/{max_retries}")
                        continue
                    else:
                        raise

        return wrapper

    return decorator


#
# Pagination and sort
#

class Pagination(BaseModel):
    """
    Pagination parameters
    """
    page: int = 1
    """ The page number """
    page_size: int = 10
    """ The page size """
    callback: Optional[Callable[[int], Awaitable[None]]] = None
    """ Callback after the search is done with the total number of hits """

    def __call__(self, model_cls: Type[TModel]) -> Type[TModel]:
        """
        Decorate the model to apply pagination

        :param model_cls: The model to decorate
        :return: The decorated model
        """

        class Wrapped(model_cls):
            """
            Decorated model class with pagination
            """

            # noinspection PyProtectedMember
            @classmethod
            async def _search(cls: model_cls, query: ESQuery, *,
                              page_size: Optional[int] = None,
                              page: int = None,
                              sort: Optional[Union[list, str]] = None,
                              **kwargs) -> ESResponse:
                page_size = self.page_size if page_size is None else page_size
                page = self.page if page is None else page
                res = await model_cls._search(query, page_size=page_size, page=page, sort=sort, **kwargs)
                if self.callback is not None:
                    try:
                        await self.callback(res['hits']['total']['value'])
                    except KeyError:
                        pass
                return res

        # Copy magic attributes
        Wrapped.__name__ = model_cls.__name__ + '_pagination'
        Wrapped.__qualname__ = model_cls.__qualname__ + '_pagination'
        Wrapped.__module__ = model_cls.__module__
        Wrapped.__doc__ = model_cls.__doc__
        Wrapped.__annotations__ = model_cls.__annotations__
        # Copy ESConfig
        Wrapped.ESConfig = model_cls.ESConfig

        return Wrapped


class SortOrder(TypedDict):
    """
    Order definition
    """
    order: Literal['asc', 'desc']


class Sort(BaseModel):
    """
    Sort parameters
    """
    sort: Union[List[Dict[str, SortOrder]], str, None]

    def __call__(self, model_cls: Type[TModel]) -> Type[TModel]:
        """
        Decorate the model to apply sort

        :param model_cls: The model to decorate
        :return: The decorated model
        """

        class Wrapped(model_cls):
            """
            Decorated model class with sort
            """

            # noinspection PyProtectedMember
            @classmethod
            async def _search(cls: model_cls, query: ESQuery, *,
                              page_size: Optional[int] = None,
                              page: int = None,
                              sort: Optional[Union[list, str]] = None,
                              **kwargs) -> ESResponse:
                sort = self.sort if sort is None else sort
                return await model_cls._search(query, page_size=page_size, page=page, sort=sort, **kwargs)

        # Copy magic attributes
        Wrapped.__name__ = model_cls.__name__ + '_sort'
        Wrapped.__qualname__ = model_cls.__qualname__ + '_sort'
        Wrapped.__module__ = model_cls.__module__
        Wrapped.__doc__ = model_cls.__doc__
        Wrapped.__annotations__ = model_cls.__annotations__
        # Copy ESConfig
        Wrapped.ESConfig = model_cls.ESConfig

        return Wrapped


#
# Index templates and mappings
#

async def create_index_template(name: str,
                                *,
                                prefix_name: str,
                                shards=1, replicas=0,
                                **other_settings: Any) -> object:
    """
    Create index template

    :param name: The name of the template
    :param prefix_name: The prefix of index pattern
    :param shards: Number of shards
    :param replicas: Number of replicas
    :param other_settings: Other settings
    :return: The result object from ES
    """
    es_version = await get_es_version()
    major, minor, _ = map(int, es_version.split('.'))

    if major > 7 or (major == 7 and minor >= 8):
        # Use composable template for ES 7.8 and above
        return await es.indices.put_index_template(
            name=name,
            body={
                "index_patterns": [f'{prefix_name}*'],
                "template": {
                    "settings": {
                        "number_of_shards": shards,
                        "number_of_replicas": replicas,
                        **other_settings
                    }
                }
            },
            request_timeout=90
        )
    else:
        # Use legacy template for ES versions below 7.8
        return await es.indices.put_template(
            name=name,
            body={
                "index_patterns": [f'{prefix_name}*'],
                "settings": {
                    "number_of_shards": shards,
                    "number_of_replicas": replicas,
                    **other_settings
                }
            },
            request_timeout=90
        )


async def setup_mappings(*_, debug=False):
    """
    Create mappings for indices or try to extend it if there are new fields
    """

    # noinspection PyShadowingNames
    def get_field_data(pydantic_type: type) -> dict:
        """ Get field data from pydantic type """
        origin = get_origin(pydantic_type)
        args = get_args(pydantic_type)

        # Handle Union type, which must be a type definition from esorm.fields, because other unions not allowed
        if origin and (
                origin is Union
                or
                getattr(origin, '__name__', None) == 'UnionType'  # UnionType is in newer Pythons, this works backwards
        ):
            # Optional may equal to Union[..., None], we don't use Optional in ES, but its child
            if type(None) in args:
                return get_field_data(args[0])

            for arg in args:
                if hasattr(arg, '__es_type__'):
                    return {'type': arg.__es_type__}

            raise ValueError('Union is not supported as ES field type!')

        # We don't use Optional in ES, but its child
        if origin is Optional:
            return get_field_data(args[0])

        # List types
        if origin is list:
            properties = {}
            create_mapping(args[0], properties)
            return {
                'type': 'nested',
                'properties': properties
            }

        # String literals
        if origin is Literal:
            return {'type': 'keyword'}

        # Pydantic annotated types
        if origin is Annotated:
            return get_field_data(args[0])

        # Origin could be a base type as well in older Python versions
        if origin in [int, float, str, bool]:
            return {'type': _pydantic_type_map[origin]}

        # Not supported origin type
        if origin:
            raise ValueError(f'Unsupported ES field type: {pydantic_type}, origin: {origin}')

        # Nested class
        if issubclass(pydantic_type, BaseModel):
            # If it is a model but has an es_type, use it (e.g. geo_point)
            if hasattr(pydantic_type, '__es_type__'):
                return {'type': pydantic_type.__es_type__}

            properties = {}
            create_mapping(pydantic_type, properties)
            return {'properties': properties}

        # Is it an ESORM type?
        if hasattr(pydantic_type, '__es_type__'):
            return {'type': pydantic_type.__es_type__}

        # Python type
        try:
            # noinspection PyTypeChecker
            return {'type': _pydantic_type_map[pydantic_type]}
        except KeyError:
            pass

        raise ValueError(f'Unknown ES field type: {pydantic_type}')

    # noinspection PyShadowingNames
    def create_mapping(model: Union[Type[BaseModel]], properties: dict):
        """ Creates mapping for the model """
        field_info: FieldInfo
        for name, field_info in model.model_fields.items():
            # Skip id field, because it won't be stored
            if hasattr(model, 'ESConfig') and model.ESConfig.id_field == name:
                continue
            extra = field_info.json_schema_extra or {}
            # Process field
            res = get_field_data(field_info.annotation)
            _type = res.get('type', None)
            if 'index' in extra and _type != 'binary':
                if 'properties' in res:
                    for v in res['properties'].values():
                        v['index'] = extra['index']
                else:
                    res['index'] = extra['index']
            properties[name] = res

    # Process all models and create mappings
    for index, model in _ESModelMeta.__models__.items():
        # Get mappings from ES if already exists
        index_exists = False
        try:
            properties = (await es.indices.get_mapping(index=index, request_timeout=90))[index]['mappings'][
                'properties']
            index_exists = True
        except (elasticsearch.NotFoundError, KeyError):
            properties = {}
        create_mapping(model, properties)

        if debug:
            from pprint import pformat
            logger.debug(
                f"`{index}` mappings:\n {pformat(properties, indent=2, width=100, compact=False, sort_dicts=False)}")

        try:
            if not index_exists:
                await es.indices.create(index=index,
                                        mappings={'properties': properties},
                                        settings=model.ESConfig.settings,
                                        request_timeout=90)
            else:
                await es.indices.put_mapping(index=index, properties=properties, request_timeout=90)
        except elasticsearch.BadRequestError:
            logger.warning(f"Index mappings error:\n{traceback.format_exc(limit=5)}")



"""
This module contains all the exceptions that can be raised by ESORM.
"""
from typing import List, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from esorm import ESModel


class NotFoundError(Exception):
    """
    Raised when a model is not found.
    """


class IndexDoesNotFoundError(Exception):
    """
    Raised when an index does not exist.
    """


class InvalidResponseError(Exception):
    """
    Raised when the response from Elasticsearch is invalid.
    """


class InvalidModelError(Exception):
    """
    Raised when a model is invalid.
    """


class BulkOperationError(TypedDict):
    """
    A dictionary type to represent an error in a bulk operation response from Elasticsearch.
    """
    # The status code of the error
    status: int
    # Type of the error
    type: str
    # A human-readable reason for the error
    reason: str
    # The model object
    model: 'ESModel'


class BulkError(Exception):
    """
    Exception for handling bulk operation errors.
    """

    failed_operations: List[BulkOperationError]

    def __init__(self, failed_operations: List[BulkOperationError]):
        super().__init__(f"Bulk operation failed for some actions: {failed_operations}")
        self.failed_operations = failed_operations



from typing import Any, Optional, Union
from functools import cached_property

from base64 import b64encode, b64decode

# noinspection PyProtectedMember
from pydantic.fields import Field as PField, FieldInfo
from pydantic_core import core_schema
from pydantic import BaseModel

__all__ = (
    'Keyword', 'Text', 'Binary', 'Byte', 'Short', 'Integer', 'Long', 'UnsignedLong', 'HalfFloat', 'Float', 'Double',
    'LatLon',
    'keyword', 'text', 'binary', 'byte', 'short', 'int32', 'long', 'unsigned_long', 'uint64', 'float16', 'float32',
    'double',
    'geo_point',
    'integer', 'half_float', 'int64', 'boolean',
    'Field', 'NumericField', 'TextField'
)


#
# Field Type Classes
#

class Keyword(str):
    """ Keyword Field """
    __es_type__ = 'keyword'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(str))


class Text(str):
    """ Text Field """
    __es_type__ = 'text'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(str))


class Binary(str):
    """
    Stores binary data as base64 encoded strings
    """
    __es_type__ = 'binary'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, __):
        return core_schema.with_info_plain_validator_function(cls.validate_binary)

    @classmethod
    def __get_pydantic_json_schema__(cls, _, handler):
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())

    @classmethod
    def validate_binary(cls, v: Union[bytes, str], _) -> str:
        if isinstance(v, bytes):
            v = b64encode(v).decode('ascii')
        return cls(v)

    @cached_property
    def bytes(self) -> bytes:
        return b64decode(self)


class Byte(int):
    """ Byte Field """
    __es_type__ = 'byte'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(int))


class Short(int):
    """ Short Field """
    __es_type__ = 'short'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(int))


class Integer(int):
    """ Integer Field """
    __es_type__ = 'integer'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(int))


class Long(int):
    """ Long Field """
    __es_type__ = 'long'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(int))


class UnsignedLong(int):
    """ Unsigned Long Field """
    __es_type__ = 'unsigned_long'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        def validate_unsigned(value: int) -> int:
            if value < 0:
                raise ValueError("Value must be an unsigned long!")
            return value

        return core_schema.no_info_after_validator_function(
            validate_unsigned,
            handler(int)
        )


class HalfFloat(float):
    """ Half Float Field """
    __es_type__ = 'half_float'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(float))


class Float(float):
    """ Float Field """
    __es_type__ = 'float'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(float))


class Double(float):
    """ Double Field """
    __es_type__ = 'double'

    @classmethod
    def __get_pydantic_core_schema__(cls, _, handler):
        return core_schema.no_info_after_validator_function(cls, handler(float))


class LatLon(BaseModel):
    """
    Geo Point Field - Latitude and Longitude
    """
    __es_type__ = 'geo_point'

    lat: float
    """Latitude Coordinate"""
    lon: float
    """Longitude Coordinate"""


#
# Field Types
#

# These types should be used in model definitions, because these can be interpreted by both Python and Pydantic

keyword = Union[Keyword, str]
""" Keyword type """
text = Union[Text, str]
""" Text type """
binary = Union[Binary, str]
""" Binary type """

byte = Union[Byte, int]
""" Byte type """
short = Union[Short, int]
""" Short type """
int32 = Union[Integer, int]
""" 32 bit integer type """
long = Union[Long, int]
""" 64 bit integer (long) type """
unsigned_long = Union[UnsignedLong, int]
""" Unsigned 64 bit integer type """

float16 = Union[HalfFloat, float]
""" 16 bit float type """
float32 = Union[Float, float]
""" 32 bit float type """
double = Union[Double, float]
""" 64 bit float (double) type """

# Aliases
integer = int32
int64 = long
uint64 = unsigned_long
boolean = bool
half_float = float16

geo_point = LatLon
""" Geo Point type """


# noinspection PyPep8Naming
def Field(
        default: Any,
        *,
        index: bool = True,
        alias: Optional[str] = None,
        # Other pydantic args
        title: Optional[str] = None,
        description: Optional[str] = None,
        exclude: Optional[bool] = None,
        include: Optional[bool] = None,
        frozen: bool = False,
        **extra
) -> FieldInfo:
    """
    Basic Field Info

    :param default: since this is replacing the field’s default, its first argument is used
        to set the default, use ellipsis (``...``) to indicate the field is required
    :param index: if this field should be indexed or not
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param exclude: exclude this field while dumping.
        Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param include: include this field while dumping.
        Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param frozen: if this field should be frozen or not
    :param extra: any additional keyword arguments will be added as is to the schema
    :return: A field info object
    """
    return PField(default, alias=alias,
                  title=title, description=description,
                  exclude=exclude, include=include, frozen=frozen,
                  index=index, json_schema_extra=extra)


# noinspection PyPep8Naming
def NumericField(
        default: Union[int, float],
        *,
        index: Optional[bool] = None,
        alias: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        allow_inf_nan: Optional[bool] = None,
        max_digits: Optional[int] = None,
        decimal_places: Optional[int] = None,
        # Other pydantic args
        title: Optional[str] = None,
        description: Optional[str] = None,
        exclude: Optional[bool] = None,
        include: Optional[bool] = None,
        frozen: bool = False,
        **extra
) -> FieldInfo:
    """
    Numeric Field Info

    :param default: since this is replacing the field’s default, its first argument is used
        to set the default, use ellipsis (``...``) to indicate the field is required
    :param index: if this field should be indexed or not
    :param alias: the public name of the field
    :param gt: only applies to numbers, requires the field to be "greater than". The schema
        will have an ``exclusiveMinimum`` validation keyword
    :param ge: only applies to numbers, requires the field to be "greater than or equal to". The
      schema will have a ``minimum`` validation keyword
    :param lt: only applies to numbers, requires the field to be "less than". The schema
        will have an ``exclusiveMaximum`` validation keyword
    :param le: only applies to numbers, requires the field to be "less than or equal to". The
        schema will have a ``maximum`` validation keyword
    :param multiple_of: only applies to numbers, requires the field to be "a multiple of". The
        schema will have a ``multipleOf`` validation keyword
    :param allow_inf_nan: only applies to numbers, allows the field to be NaN or infinity (+inf or -inf),
        which is a valid Python float. Default True, set to False for compatibility with JSON.
    :param max_digits: only applies to Decimals, requires the field to have a maximum number
        of digits within the decimal. It does not include a zero before the decimal point or trailing decimal zeroes.
    :param decimal_places: only applies to Decimals, requires the field to have at most a number of decimal places
        allowed. It does not include trailing decimal zeroes.
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param exclude: exclude this field while dumping.
        Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param include: include this field while dumping.
        Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param frozen: if this field should be frozen or not
    :param extra: any additional keyword arguments will be added as is to the schema
    :return: A field info object
    """
    extra = dict(extra)
    if index is not None:
        extra['index'] = index
    return PField(default, alias=alias,
                  gt=gt, ge=ge, lt=lt, le=le,
                  multiple_of=multiple_of, allow_inf_nan=allow_inf_nan,
                  max_digits=max_digits, decimal_places=decimal_places,
                  title=title, description=description,
                  exclude=exclude, include=include, frozen=frozen,
                  json_schema_extra=extra)


# noinspection PyPep8Naming
def TextField(
        default: str,
        *,
        index: bool = True,
        alias: Optional[str] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        # Other pydantic args
        title: Optional[str] = None,
        description: Optional[str] = None,
        exclude: Optional[bool] = None,
        include: Optional[bool] = None,
        frozen: bool = False,
        **extra
) -> FieldInfo:
    """
    Text Field Info

    :param default: since this is replacing the field’s default, its first argument is used
        to set the default, use ellipsis (``...``) to indicate the field is required
    :param index: if this field should be indexed or not
    :param alias: the public name of the field
    :param min_length: only applies to strings, requires the field to have a minimum length. The
        schema will have a ``minLength`` validation keyword
    :param max_length: only applies to strings, requires the field to have a maximum length. The
        schema will have a ``maxLength`` validation keyword
    :param regex: only applies to strings, requires the field match against a regular expression
        pattern string. The schema will have a ``pattern`` validation keyword
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param exclude: exclude this field while dumping.
        Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param include: include this field while dumping.
        Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param frozen: if this field should be frozen or not
    :param extra: any additional keyword arguments will be added as is to the schema
    :return: A field info object
    """
    extra = dict(extra)
    if index is not None:
        extra['index'] = index
    return PField(default, alias=alias,
                  min_length=min_length, max_length=max_length,
                  regex=regex,
                  title=title, description=description,
                  exclude=exclude, include=include, frozen=frozen,
                  json_schema_extra=extra)



"""
ElasticSearch Watcher support for ESORM
"""
from typing import TypedDict, Tuple, Dict, List, Any, Union, Type, Optional, cast

import json

from .logger import logger
from .query import ESQuery
from .esorm import es


class WatcherMeta(type):
    """
    Watcher metaclass
    """
    __watchers__: List[Type['Watcher']] = []

    def __init__(cls: Type['Watcher'], name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]) -> None:
        super().__init__(name, bases, attrs)

        WatcherMeta.__watchers__.append(cls)
        for base in bases:
            base = cast(Type['Watcher'], base)
            if base in WatcherMeta.__watchers__:
                WatcherMeta.__watchers__.remove(base)


class Schedule(TypedDict):
    """ Schedule definition """
    interval: str


class Trigger(TypedDict):
    """ Trigger definition """
    schedule: Schedule


class Order(TypedDict):
    """ Order definition """
    order: str


class Body(TypedDict, total=False):
    """ Body definition """
    size: int
    sort: Dict[str, Order]
    query: ESQuery


class Request(TypedDict, total=False):
    """ Request definition """
    indices: Union[List[str], str]
    body: ESQuery
    tepmlate: Dict[str, Any]


class Search(TypedDict, total=False):
    """ Search definition """
    request: Request
    extract: List[str]


class EmptyDict(TypedDict, total=False):
    """ Empty dict definition """


class Compare(TypedDict, total=False):
    """ Compare definition """
    eq: Any
    not_eq: Any
    gt: Union[int, float, str, Dict[str, Any]]
    gte: Union[int, float, str, Dict[str, Any]]
    lt: Union[int, float, str, Dict[str, Any]]
    lte: Union[int, float, str, Dict[str, Any]]


class ArrayCompare(TypedDict, total=False):
    """ Array compare definition """
    path: str
    eq: Any
    not_eq: Any
    gt: Union[int, float, str, Dict[str, Any]]
    gte: Union[int, float, str, Dict[str, Any]]
    lt: Union[int, float, str, Dict[str, Any]]
    lte: Union[int, float, str, Dict[str, Any]]


class Condition(TypedDict, total=False):
    """ Condition definition """
    compare: Dict[str, Compare]
    array_compare: Dict[str, Any]
    never: EmptyDict
    always: EmptyDict
    script: Dict[str, Any]


class Transform(TypedDict, total=False):
    """ Transform definition """
    search: Dict[str, Any]
    script: Dict[str, Any]
    chain: List[Dict[str, Any]]


class ActionWebhook(TypedDict, total=False):
    """ Action webhook definition """
    scheme: str
    host: str
    port: int
    method: str
    path: str
    params: Dict[str, Any]
    headers: Dict[str, Any]
    body: str
    auth: Dict[str, Any]
    proxy: Dict[str, Any]
    timeout: str
    connection_timeout: str
    read_timeout: str
    retries: int
    retry_on_status: List[int]
    ssl: Dict[str, Any]
    webhook: Dict[str, Any]


class Action(TypedDict, total=False):
    """ Action definition """
    transform: Dict[str, Transform]
    throttle_period: str

    email: Dict[str, Any]
    index: Dict[str, Any]
    logging: Dict[str, Any]
    pagerduty: Dict[str, Any]
    slack: Dict[str, Any]
    webhook: ActionWebhook


class Watcher(metaclass=WatcherMeta):
    """
    Watcher definition
    """
    metadata: Optional[Dict[str, Any]] = None
    trigger: Optional[Trigger] = None
    input: Optional[Search] = None
    condition: Optional[Condition] = None
    actions: Optional[Dict[str, Action]] = None

    def __init__(self):
        assert self.trigger is not None, "Trigger is not defined"

    def to_es(self):
        res = {}
        if self.metadata is not None:
            res['metadata'] = self.metadata
        if self.trigger is not None:
            res['trigger'] = self.trigger
        if self.input is not None:
            res['input'] = self.input
        if self.condition is not None:
            res['condition'] = self.condition
        if self.actions is not None:
            res['actions'] = self.actions
        return res


class DeleteWatcher(Watcher):
    """
    Watcher for deleting documents matching a query
    """
    _scheme: str = "http"
    """ ES scheme """
    _host: str = "127.0.0.1"
    """ ES host """
    _port: int = 9200
    """ ES port """

    _index: Optional[str] = None
    """ Target index """
    _query: Optional[ESQuery] = None
    """ Query to match documents to delete """

    def __init__(self):
        super().__init__()

        assert self._index is not None, "Index is not defined"
        assert self._query is not None, "Query is not defined"

        self.actions = {
            "delete_doc": {
                "webhook": {
                    "scheme": self._scheme,
                    "method": "POST",
                    "host": self._host,
                    "port": self._port,
                    "path": f"/{self._index}/_delete_by_query",
                    "body": f'{{"query": {json.dumps(self._query)}}}',
                }
            }
        }


async def setup_watchers(*_, debug=False):
    """
    Setup watchers
    :param _: Unused
    :param debug: Whether to print the watcher definition
    """
    for WatcherClass in WatcherMeta.__watchers__:
        watcher = WatcherClass()
        if debug:
            from pprint import pformat
            logger.debug(
                f"`{Watcher.__name__}` watcer:\n {pformat(watcher.to_es(), indent=2, width=100, compact=False, sort_dicts=False)}")
        await es.watcher.put_watch(id=Watcher.__name__, **watcher.to_es())
        logger.info(f"Watcher {Watcher.__name__} created.")



"""
This module contains type definitions for the response from Elasticsearch.
"""
from typing import TypedDict, List, Dict, Union, Any
from . import aggs


class Hit(TypedDict):
    """
    Represents a single hit (result) from Elasticsearch.
    """
    _index: str
    """ The index the hit is from. """
    _type: str
    """ The type of the hit. """
    _id: str
    """ The id of the hit. """
    _score: Union[float, None]
    """ The score of the hit. """
    _source: Dict[str, Any]
    """ The source of the hit. """


class Hits(TypedDict):
    """
    Represents the hits section of the Elasticsearch response.
    """
    total: Dict[str, int]
    """ The total number of hits. """
    max_score: Union[float, None]
    """ The maximum score of the hits. """
    hits: List[Hit]
    """ List of hits. """


class ESResponse(TypedDict):
    """
    Represents the overall structure of an Elasticsearch response.
    """
    took: int
    """ The time in milliseconds it took to execute the query. """
    timed_out: bool
    """ Whether the query timed out. """
    _shards: Dict[str, int]
    """ The number of shards the query was executed on. """
    hits: Hits
    """ The hits section of the response. """
    aggregations: aggs.ESAggsResponse
    """ The aggregations section of the response. """



"""
FastaAPI utilities for ESORM
"""
from typing import Union, List, Dict

import inspect
import io
import tokenize

from enum import Enum

from fastapi import Response

from .model import Pagination, Sort
from .utils import camel_case

_max_page_size = 10000

__all__ = [
    'make_dep_pagination',
    'make_dep_sort',
    'set_max_page_size',
]


#
# Pagination and sort dependencies
#

def set_max_page_size(max_page_size: int):
    """
    Set the maximum page size for queries

    :param max_page_size: The maximum page size
    :return: None
    """
    global _max_page_size
    _max_page_size = max_page_size


def make_dep_pagination(default_page: int = 1, default_page_size: int = 10, set_headers: bool = True) -> callable:
    """
    Create a pagination dependency with default values

    :param default_page: Default page number, the first page is 1
    :param default_page_size: Default page size, the default is 10
    :param set_headers: Set X-Total-Hits header after search
    :return: Pagination dependency
    """

    async def _dep_pagination(response: Response,
                              _page: int = default_page, _page_size: int = default_page_size) -> Pagination:
        """
        Pagination dependency

        :param response: Response object
        :param _page: Page number, the first page is 1
        :param _page_size: Page size, the default is 10
        :return: Pagination object
        """

        async def _cb_search(total_hits: int):
            """
            Callback function after search with total hits

            :param total_hits: Total hits of the query
            :return:
            """
            # NOTE: it should be in the exposed_headers in CORS middleware!!
            response.headers['X-Total-Hits'] = str(total_hits)

        # Ensure page size is not too large
        _page_size = min(_page_size, max(_max_page_size, default_page_size))
        return Pagination(page=_page, page_size=_page_size, callback=_cb_search if set_headers else None)

    return _dep_pagination


def make_dep_sort(**kwargs: Union[List[Dict[str, dict]], Dict[str, any]]) -> callable:
    """
    Create a sort dependency with sort definitions

    :param kwargs: Sort definitions
    :return: Sort dependency
    """
    assert len(kwargs) > 0, "At least one sort field is required"

    # Get the name of the function where this function is called
    frame = inspect.currentframe().f_back
    module_name = inspect.getmodule(frame).__name__
    source_code = inspect.getsource(frame)
    tokens = list(tokenize.tokenize(io.BytesIO(source_code.encode('utf-8')).readline))
    function_name = ""
    for i, token in enumerate(tokens):
        if token.type == tokenize.NAME and token.string == "def" and token.start[0] <= frame.f_lineno:
            function_name = tokens[i + 1].string
        elif token.type == tokenize.NAME and token.string == "def" and token.start[0] > frame.f_lineno:
            break
    assert function_name, "Cannot find function name!"
    fqdn = f"{module_name}.{function_name}"
    # Create a new enum class for sort fields
    sort_enum = Enum('SortEnum' + camel_case(fqdn.replace('.', '_'), True),
                     {k: k for k in kwargs.keys()})

    async def _dep_sort(_sort: sort_enum = None) -> Sort:
        """
        Sort dependency

        :param _sort: Sort JSON string
        :return: Sort object
        """
        return Sort(sort=kwargs[_sort.name] if _sort else None)

    return _dep_sort



"""
Bulk operation for ElasticSearch
"""
from .error import BulkError, BulkOperationError
from .utils import utcnow

from .model import TModel, ESModelTimestamp
from .esorm import es


class ESBulk:
    """
    Bulk operation for ElasticSearch
    """

    def __init__(self, wait_for=False, **bulk_kwargs):
        """
        Create a bulk context manager

        :param wait_for: Whether to wait for active shards
        :param bulk_kwargs: Other bulk arguments
        """
        self._actions = []
        self._bulk_kwargs = dict(bulk_kwargs)
        self._models_to_index = []
        if wait_for:
            self._bulk_kwargs['refresh'] = 'wait_for'

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Do the bulk operation
        :param exc_type: Exception type
        :param exc_val: Exception value
        :param exc_tb: Exception traceback
        :return
        """
        # Do the bulk operation if no exception is raised
        if exc_val is None:
            res = await es.bulk(operations=self._actions, **self._bulk_kwargs)
            errors = []
            for idx, item in enumerate(res['items']):
                model = self._models_to_index[idx]
                for action, result in item.items():
                    model._id = result.get('_id', None)

                    if 'error' in result:
                        error: BulkOperationError = {
                            'status': result['status'],
                            'type': result['error']['type'],
                            'reason': result['error']['reason'],
                            'model': model
                        }
                        errors.append(error)

                    elif action == 'index':
                        model._seq_no = result['_seq_no']
                        model._primary_term = result['_primary_term']
                        model._version = result['_version']
            if errors:
                raise BulkError(errors)

        # The exceptions are not handled here, propagate them
        return False

    # noinspection PyProtectedMember
    async def save(self, model: TModel):
        """
        Add the model to the bulk for saving

        If the model is from ES (get or search, so it has _seq_no and _primary_term), it will
        use optimistic concurrency check, so it will only update the document if the _seq_no and
        _primary_term are the same as the document in the index.

        If the model is an ESModelTimestamp, it will update the modified_at field to the current
        time and if the created_at field is not already set, it will set it to the current time too.

        :param model: The model to add for saving
        """
        document: dict = model.to_es()
        if model.ESConfig.id_field:
            del document[model.ESConfig.id_field]

        index = {
            '_index': model.ESConfig.index,
            '_id': model.__id__
        }
        routing = model.__routing__
        if routing is not None:
            index['routing'] = routing

        # Optimistic concurrency check
        if model._primary_term and model._seq_no:
            index['if_primary_term'] = model._primary_term
            index['if_seq_no'] = model._seq_no

        # Support for ESModelTimestamp
        if isinstance(model, ESModelTimestamp):
            document['modified_at'] = utcnow()
            # Support for created_at field
            if not model.created_at:
                document['created_at'] = document['modified_at']

        # Save model for later to update private fields after the bulk operation
        self._models_to_index.append(model)

        # Add action and document to the bulk
        self._actions.append({'index': index})
        self._actions.append(document)

    async def delete(self, model: TModel):
        """
        Add the model to the bulk for deletion

        :param model: The model to add for deletion
        """
        delete = {
            '_index': model.ESConfig.index,
            '_id': model.__id__
        }
        routing = model.__routing__
        if routing is not None:
            delete['routing'] = routing

        # Save model for later to update private fields after the bulk operation
        self._models_to_index.append(model)

        # Add action
        self._actions.append({'delete': delete})



"""
Utility functions
"""
import re
import datetime


def snake_case(camel_str: str):
    """
    Convert to snake case

    :param camel_str: The string to convert to snake case
    :return: Converted string
    """
    if '_' in camel_str:  # If it is already snake cased
        return camel_str
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', camel_str)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def camel_case(snake_str: str, capitalize_first: bool = False):
    """
    Convert to camel case

    :param snake_str: The string to convert to camel case
    :param capitalize_first: Capitalize the first letter
    :return: Converted string
    """
    components = snake_str.split('_')
    if capitalize_first:
        components[0] = components[0].title()
    return components[0] + ''.join((x.title()) for x in components[1:])


def utcnow():
    """
    Get current UTC time

    :return: Current UTC time
    """
    return datetime.datetime.now(datetime.timezone.utc)



import asyncio
from typing import TYPE_CHECKING
import sys
import pytest


@pytest.mark.parametrize('service', ['es8x', 'es7x'], scope='class')
class TestBaseTests:
    """
    Base tests
    """

    doc_id: str = None

    async def test_connect(self, es):
        """
        Test connection
        """
        assert es is not None

    async def test_create_basemodel(self, es, esorm):
        """
        Test create base model
        """

        class BaseModel(esorm.ESBaseModel):
            f_str: str
            f_int: int

        class BaseChild(BaseModel):
            f_float: float

        # Basemodels should not be registered as models they should not be created in the DB as indexes
        # TODO: This is not working in the 2nd run, this is only test issue
        # assert len(esorm.model._ESModelMeta.__models__) == 0
        assert not hasattr(BaseModel.ESConfig, 'index')
        assert not hasattr(BaseChild.ESConfig, 'index')

    async def test_create_model_with_python_fields(self, es, esorm, model_python):
        """
        Test model with python fields
        """
        assert model_python is not None
        assert model_python.ESConfig.index == 'esorm_python_field_model'

    async def test_create_model_with_es_fields(self, es, esorm, model_es):
        """
        Test model with ES fields
        """
        assert model_es is not None
        assert model_es.ESConfig.index == 'esorm_es_field_model'

    async def test_create_model_with_es_optional_fields(self, es, esorm, model_es_optional):
        """
        Test model with ES fields
        """
        assert model_es_optional is not None
        assert model_es_optional.ESConfig.index == 'esorm_es_optional_field_model'

    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10 or higher")
    async def test_create_model_with_es_optional_fields_new_syntax(self, es, esorm, model_es_optional_new_syntax):
        """
        Test model with ES fields with new syntax
        """
        assert model_es_optional_new_syntax is not None
        assert model_es_optional_new_syntax.ESConfig.index == 'esorm_es_optional_field_new_syntax_model'

    async def test_create_timestamp_models(self, es, esorm, model_timestamp):
        """
        Test timestamp models
        """
        assert model_timestamp is not None
        assert model_timestamp.ESConfig.index == 'esorm_timestamp_model'
        assert 'created_at' in model_timestamp.model_fields
        assert 'modified_at' in model_timestamp.model_fields

    async def test_create_model_config(self, es, esorm, model_config):
        """
        Test model config
        """
        assert model_config is not None
        assert model_config.ESConfig.index == 'custom_index'

    async def test_model_template(self, es, esorm, model_index_template):
        """
        Test model config
        """
        assert model_index_template is not None
        assert model_index_template.ESConfig.index == 'custom_index_template'

    async def test_create_model_with_id(self, es, esorm, model_with_id):
        """
        Test model with config
        """
        assert model_with_id is not None
        assert model_with_id.ESConfig.index == 'esorm_id_model'
        assert model_with_id.ESConfig.id_field == 'id'

    async def test_create_model_with_int_id(self, es, esorm, model_with_int_id):
        """
        Test model with int id
        """
        assert model_with_int_id is not None
        assert model_with_int_id.ESConfig.index == 'esorm_int_id_model'
        assert model_with_int_id.ESConfig.id_field == 'custom_id'

    async def test_create_model_with_prop_id(self, es, esorm, model_with_prop_id):
        """
        Test model with property id
        """
        assert model_with_prop_id is not None
        assert model_with_prop_id.ESConfig.index == 'esorm_prop_id_model'
        assert model_with_prop_id.ESConfig.id_field is None

    async def test_create_nested_model(self, es, esorm, model_nested):
        """
        Test nested model
        """
        assert model_nested is not None
        assert model_nested.ESConfig.index == 'esorm_nested_field_model'

    async def test_create_mappings(self, es, esorm, model_python, model_es,
                                   model_es_optional, model_es_optional_new_syntax,
                                   model_timestamp, model_config, model_with_id, model_with_int_id,
                                   model_with_prop_id, model_nested, model_lazy_prop,
                                   model_index_template):
        """
        Test create mappings
        """
        # Create index template
        await esorm.model.create_index_template('custom_index_template', prefix_name='custom_index_',
                                                shards=5, auto_expand_replicas="1-2")
        await esorm.setup_mappings()

        # Check if index exists
        assert await es.indices.exists(index=model_python.ESConfig.index)

        # Check if mappings are correct for python fields
        mappings = await es.indices.get_mapping(index=model_python.ESConfig.index)
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_str']['type'] == 'keyword'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_int']['type'] == 'long'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_float']['type'] == 'double'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_bool']['type'] == 'boolean'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_datetime']['type'] == 'date'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_date']['type'] == 'date'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_time']['type'] == 'date'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_literal']['type'] == 'keyword'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_uuid4']['type'] == 'keyword'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_file_path']['type'] == 'keyword'
        assert mappings[model_python.ESConfig.index]['mappings']['properties']['f_http_url']['type'] == 'keyword'

        # Check if mappings are correct for ES fields
        mappings = await es.indices.get_mapping(index=model_es.ESConfig.index)
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_keyword']['type'] == 'keyword'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_text']['type'] == 'text'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_binary']['type'] == 'binary'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_byte']['type'] == 'byte'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_short']['type'] == 'short'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_int32']['type'] == 'integer'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_long']['type'] == 'long'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_unsigned_long']['type'] == 'unsigned_long'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_float16']['type'] == 'half_float'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_float32']['type'] == 'float'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_double']['type'] == 'double'
        assert mappings[model_es.ESConfig.index]['mappings']['properties']['f_geo_point']['type'] == 'geo_point'

        # Check if mappings are correct for ES Optional fields
        mappings = await es.indices.get_mapping(index=model_es_optional.ESConfig.index)
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_keyword']['type'] == 'keyword'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_text']['type'] == 'text'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_binary']['type'] == 'binary'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_byte']['type'] == 'byte'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_short']['type'] == 'short'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_int32']['type'] == 'integer'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_long']['type'] == 'long'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_unsigned_long'][
                   'type'] == 'unsigned_long'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_float16']['type'] == 'half_float'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_float32']['type'] == 'float'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_double']['type'] == 'double'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_geo_point'][
                   'type'] == 'geo_point'
        assert mappings[model_es_optional.ESConfig.index]['mappings']['properties']['f_positive_int']['type'] == 'long'

        # Test optional new syntax
        if sys.version_info >= (3, 10):
            mappings = await es.indices.get_mapping(index=model_es_optional_new_syntax.ESConfig.index)
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_keyword'][
                       'type'] == 'keyword'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_text'][
                       'type'] == 'text'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_binary'][
                       'type'] == 'binary'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_byte'][
                       'type'] == 'byte'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_short'][
                       'type'] == 'short'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_int32'][
                       'type'] == 'integer'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_long'][
                       'type'] == 'long'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_unsigned_long'][
                       'type'] == 'unsigned_long'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_float16'][
                       'type'] == 'half_float'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_float32'][
                       'type'] == 'float'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_double'][
                       'type'] == 'double'
            assert mappings[model_es_optional_new_syntax.ESConfig.index]['mappings']['properties']['f_geo_point'][
                       'type'] == 'geo_point'

        # Check if mappings are correct for timestamp fields
        mappings = await es.indices.get_mapping(index=model_timestamp.ESConfig.index)
        assert mappings[model_timestamp.ESConfig.index]['mappings']['properties']['created_at']['type'] == 'date'
        assert mappings[model_timestamp.ESConfig.index]['mappings']['properties']['modified_at']['type'] == 'date'

        # Check if mappings are correct for model with config
        mappings = await es.indices.get_mapping(index=model_config.ESConfig.index)
        assert mappings[model_config.ESConfig.index]['mappings']['properties']['f_str']['type'] == 'keyword'
        # Id should not be in mappings
        assert 'custom_id' not in mappings[model_config.ESConfig.index]['mappings']['properties'], \
            "Id fields should not be in mappings"
        # Check index settings
        settings = await es.indices.get_settings(index=model_config.ESConfig.index)
        assert settings[model_config.ESConfig.index]['settings']['index']['number_of_shards'] == '6'
        assert settings[model_config.ESConfig.index]['settings']['index']['number_of_replicas'] == '1'
        assert settings[model_config.ESConfig.index]['settings']['index']['refresh_interval'] == '5s'

        # Check if mappings are correct for model with index template
        mappings = await es.indices.get_mapping(index=model_index_template.ESConfig.index)
        assert mappings[model_index_template.ESConfig.index]['mappings']['properties']['f_str']['type'] == 'keyword'
        # Check index settings
        settings = await es.indices.get_settings(index=model_index_template.ESConfig.index)
        assert settings[model_index_template.ESConfig.index]['settings']['index']['number_of_shards'] == '5'
        assert settings[model_index_template.ESConfig.index]['settings']['index']['auto_expand_replicas'] == '1-2'

        # Check if mappings are correct for model with id
        mappings = await es.indices.get_mapping(index=model_with_id.ESConfig.index)
        assert mappings[model_with_id.ESConfig.index]['mappings']['properties']['f_str']['type'] == 'keyword'
        assert 'id' not in mappings[model_with_id.ESConfig.index]['mappings']['properties'], \
            "Id fields should not be in mappings"

        # Check if mappings are correct for model with int id
        mappings = await es.indices.get_mapping(index=model_with_int_id.ESConfig.index)
        assert mappings[model_with_int_id.ESConfig.index]['mappings']['properties']['f_str']['type'] == 'keyword'
        assert 'custom_id' not in mappings[model_with_int_id.ESConfig.index]['mappings']['properties'], \
            "Id fields should not be in mappings"

        # Check if mappings are correct for model with prop id
        mappings = await es.indices.get_mapping(index=model_with_prop_id.ESConfig.index)
        assert mappings[model_with_prop_id.ESConfig.index]['mappings']['properties']['f_str']['type'] == 'keyword'
        # It is used in the id, but remains in the db
        assert mappings[model_with_prop_id.ESConfig.index]['mappings']['properties']['custom_id']['type'] == 'long'

        # Check if mappings are correct for nested model
        mappings = await es.indices.get_mapping(index=model_nested.ESConfig.index)
        assert 'properties' in mappings[model_nested.ESConfig.index]['mappings']['properties']['f_nested'], \
            "Nested field should have properties"
        assert 'created_at' in mappings[model_nested.ESConfig.index]['mappings']['properties']['f_nested'][
            'properties']
        assert 'modified_at' in mappings[model_nested.ESConfig.index]['mappings']['properties']['f_nested'][
            'properties']
        assert 'f_str' in mappings[model_nested.ESConfig.index]['mappings']['properties']['f_nested'][
            'properties']

    async def test_crud_create(self, es, esorm, model_nested, model_timestamp):
        """
        Test create
        """
        # It should have a ES generated document id
        doc_id = await model_nested(f_nested=model_timestamp(f_str="nested_test")).save()
        assert doc_id is not None and len(doc_id) == 20, "Incorrect document id"
        self.__class__.doc_id = doc_id  # Strange but it works

        # Create some other documents
        doc_id = await model_nested(f_nested=model_timestamp(f_str="nested_test2", f_int=2), f_float=1.0).save()
        assert doc_id is not None and len(doc_id) == 20, "Incorrect document id"
        doc_id = await model_nested(f_nested=model_timestamp(f_str="nested_test3", f_int=3), f_float=1.5).save()
        assert doc_id is not None and len(doc_id) == 20, "Incorrect document id"
        doc_id = await model_nested(f_nested=model_timestamp(f_str="nested_test4", f_int=4), f_float=2.0).save()
        assert doc_id is not None and len(doc_id) == 20, "Incorrect document id"
        doc_id = await model_nested(f_nested=model_timestamp(f_str="nested_test5", f_int=5), f_float=2.5).save()
        assert doc_id is not None and len(doc_id) == 20, "Incorrect document id"

    async def test_crud_get_by_id(self, es, esorm, model_nested, model_timestamp):
        """
        Test get by id
        """
        doc_id = self.__class__.doc_id
        # Get by id
        doc = await model_nested.get(doc_id)
        assert doc is not None
        assert doc.f_float == 0.5  # Default value of the field
        assert doc.f_nested is not None
        assert doc.f_nested.f_str == "nested_test"

        # Check private fields
        assert doc._version == 1
        assert doc._primary_term == 1
        assert doc._seq_no == 0

    async def test_crud_get_by_int_id(self, es, esorm, model_with_int_id):
        """
        Test get by int id
        """
        # Create a document
        doc_id = await model_with_int_id(custom_id=1, f_str="int_id_test").save()
        assert doc_id == "1"

        # Get by id
        doc = await model_with_int_id.get(1)
        assert doc is not None
        assert doc.custom_id == 1
        assert doc.f_str == "int_id_test"
        assert doc._id == "1"

    async def test_crud_get_by_prop_id(self, es, esorm, model_with_prop_id):
        """
        Test get by prop id
        """
        # Create a document
        doc_id = await model_with_prop_id(custom_id=1, f_str="prop_id_test").save()
        assert doc_id == "1001"

        # Get by id
        doc = await model_with_prop_id.get(1001)
        assert doc is not None
        assert doc.custom_id == 1
        assert doc.f_str == "prop_id_test"
        assert doc._id == "1001"

    async def test_crud_update(self, es, esorm, model_nested):
        """
        Test update
        """
        doc_id = self.__class__.doc_id
        # Update
        doc = await model_nested.get(doc_id)
        modified_at = doc.f_nested.modified_at
        created_at = doc.f_nested.created_at
        assert doc is not None
        doc.f_nested.f_str = "nested_test_updated"
        await doc.save()

        await asyncio.sleep(0.1)

        # Get by id
        doc = await model_nested.get(doc_id)
        assert doc is not None
        assert doc.f_nested.f_str == "nested_test_updated"
        # Test if modified_at is updated even in nested fields
        assert doc.f_nested.modified_at > modified_at
        # Test if created_at is not updated
        assert doc.f_nested.created_at == created_at

        # Check private fields
        assert doc._version == 2
        assert doc._primary_term == 1
        assert doc._seq_no == 5  # It is 5 because other documents have been created

    async def test_python_types(self, es, esorm, model_python):
        """
        Test pydantic types
        """
        from datetime import datetime, date, time
        from uuid import uuid4, UUID

        await esorm.setup_mappings()

        doc = model_python(
            f_str='test', f_int=123, f_float=0.123, f_bool=True,
            f_datetime=datetime.now(),
            f_date=date.today(), f_time=datetime.now().time(),
            f_literal='a',
            f_uuid4=uuid4(),
            f_file_path='/bin/sh',
            f_http_url='http://example.com'
        )
        doc_id = await doc.save()
        assert doc_id is not None

        # Get by id
        doc = await model_python.get(doc_id)
        assert doc is not None
        assert doc.f_str == 'test'
        assert doc.f_int == 123
        assert doc.f_float == 0.123
        assert doc.f_bool is True
        assert isinstance(doc.f_datetime, datetime)
        assert isinstance(doc.f_date, date)
        assert isinstance(doc.f_time, time)
        assert doc.f_literal == 'a'
        assert isinstance(doc.f_uuid4, UUID)
        assert str(doc.f_file_path) == '/bin/sh'
        assert str(doc.f_http_url) == 'http://example.com/'

    async def test_race_condition(self, es, esorm, model_nested):
        """
        Test race condition
        Race conditions should raise a ConflictError
        """
        from elasticsearch import ConflictError

        ### Test update ###

        doc_id = self.__class__.doc_id
        # Get 2 instances of the same document
        doc1 = await model_nested.get(doc_id)
        doc1.f_float = 1.0
        doc2 = await model_nested.get(doc_id)
        doc1.f_float = 2.0

        # The 2nd save should not work
        with pytest.raises(ConflictError):
            await doc1.save()
            await doc2.save()

        ### Test delete ###

        # Get 2 instances of the same document
        doc1 = await model_nested.get(doc_id)
        doc1.f_float = 3.0
        doc2 = await model_nested.get(doc_id)

        # Update and delete at the same time
        with pytest.raises(ConflictError):
            await doc1.save(wait_for=True)
            await doc2.delete()

        doc = await model_nested.get(doc_id)
        assert doc is not None
        assert doc.f_float == 3.0

    async def test_retry_on_conflict(self, es, esorm, model_timestamp):
        """
        Test optimistic concurrency control with retry
        """
        import asyncio
        from elasticsearch import ConflictError

        doc = model_timestamp(f_str="occ_retry", f_int=10)
        await doc.save()

        @esorm.retry_on_conflict(3)
        async def update_doc(doc_id):
            _doc = await model_timestamp.get(doc_id)
            _doc.f_int += 1
            await _doc.save()

        await asyncio.gather(
            update_doc(doc._id),
            update_doc(doc._id),
            update_doc(doc._id),
        )

        await doc.reload()
        assert doc.f_int == 13

        doc = model_timestamp(f_str="occ_retry", f_int=10)
        await doc.save()

        # Test it without retry logic

        async def update_without_retry(doc_id):
            _doc = await model_timestamp.get(doc_id)
            _doc.f_int += 1
            await _doc.save()

        with pytest.raises(ConflictError):
            await asyncio.gather(
                update_without_retry(doc._id),
                update_without_retry(doc._id),
                update_without_retry(doc._id),
            )

    # noinspection PyBroadException
    async def test_crud_delete(self, es, esorm, model_nested):
        """
        Test delete
        """
        doc_id = self.__class__.doc_id
        # Delete
        doc = await model_nested.get(doc_id)
        assert doc is not None
        await doc.delete()

        # Get by id
        try:
            await model_nested.get(doc_id)
            assert False, "Document should not exist"
        except Exception as e:
            assert e.__class__.__name__ == 'NotFoundError', "Incorrect exception"
            self.__class__.doc_id = None

    async def test_bulk_operations(self, es, esorm, model_nested, model_timestamp):
        """
        Test bulk operations
        """
        # Creating documents
        async with esorm.ESBulk(wait_for=True) as bulk:  # Here wait_for is important!
            for i in range(10):
                doc = model_nested(f_nested=model_timestamp(f_str=f"nested_test1{i}", f_int=10 + i), f_float=10.0 + i)
                await bulk.save(doc)

        # Deleting documents
        async with esorm.ESBulk(wait_for=True) as bulk:  # Here wait_for is important!
            for i in range(10):
                doc = await model_nested.search_one_by_fields({'f_nested.f_str': f"nested_test1{i}"})
                assert doc is not None
                await bulk.delete(doc)

    async def test_bulk_race_condition_and_reload(self, es, esorm, model_nested, model_timestamp):
        """
        Test bulk operations with conflict
        """
        doc = model_nested(f_nested=model_timestamp(f_str=f"nested_test1"), f_float=10.0)
        await doc.save()

        with pytest.raises(esorm.error.BulkError):
            async with esorm.ESBulk(wait_for=True) as bulk:
                doc1 = await model_nested.search_one_by_fields({'f_nested.f_str': f"nested_test1"})
                assert doc1._version == 1
                doc2 = await model_nested.search_one_by_fields({'f_nested.f_str': f"nested_test1"})
                assert doc2._version == 1
                assert doc1 is not None
                assert doc2 is not None
                doc1.f_float = 11.0
                doc1.f_nested.f_str = "nested_test1_updated1"
                await bulk.save(doc1)
                # This should not work, because this is a race condition
                doc2.f_float = 11.1
                doc1.f_nested.f_str = "nested_test1_updated1"
                await bulk.save(doc2)

        # We test reloading here, it should now update the _primary_term and _seq_no
        await doc.reload()
        # So the deletion of document should be ok
        await doc.delete()

    async def test_search(self, es, esorm, model_nested):
        """
        Test search
        """
        if TYPE_CHECKING:
            from esorm.query import ESQuery

        # Search
        query: 'ESQuery' = {
            'bool': {
                'must': {
                    'match': {
                        'f_nested.f_str': 'nested_test2'
                    }
                }
            }
        }
        docs = await model_nested.search(query)
        assert len(docs) == 1
        assert docs[0].f_nested.f_str == 'nested_test2'

        # Should include private fields
        assert docs[0]._version == 1
        assert docs[0]._primary_term == 1
        assert docs[0]._seq_no > 0

        # Search all started with nested_test
        query: 'ESQuery' = {
            'bool': {
                'must': {
                    'prefix': {
                        'f_nested.f_str': 'nested_test'
                    }
                }
            }
        }
        docs = await model_nested.search(query)
        assert len(docs) == 4
        for doc in docs:
            assert doc.f_nested.f_str.startswith('nested_test')

    async def test_search_one(self, es, esorm, model_nested):
        """
        Test search one
        """
        if TYPE_CHECKING:
            from esorm.query import ESQuery

        # Search one
        query: 'ESQuery' = {
            'bool': {
                'must': {
                    'match': {
                        'f_nested.f_str': 'nested_test2'
                    }
                }
            }
        }
        doc = await model_nested.search_one(query)
        assert doc is not None
        assert doc.f_nested.f_str == 'nested_test2'

        # Search one with multiple results
        query: 'ESQuery' = {
            'bool': {
                'must': {
                    'prefix': {
                        'f_nested.f_str': 'nested_test'
                    }
                }
            }
        }
        doc = await model_nested.search_one(query)
        assert doc is not None

    async def test_search_by_fields(self, es, esorm, model_nested):
        """
        Test search by fields
        """
        # Search by fields
        docs = await model_nested.search_by_fields({
            'f_nested.f_str': 'nested_test2'
        })
        assert len(docs) == 1
        assert docs[0].f_nested.f_str == 'nested_test2'

        # Search by fields with multiple results
        doc = await model_nested.search_one_by_fields({
            'f_nested.f_str': 'nested_test3'
        })
        assert doc is not None
        assert doc.f_nested.f_str == 'nested_test3'

    async def test_basic_aggregations(self, es, esorm, model_nested):
        """
        Test basic aggregations
        """
        if TYPE_CHECKING:
            from esorm.query import ESQuery
            from esorm.aggs import ESAggsResponse, ESAggs

        # Test aggregations without query
        aggs_def: 'ESAggs' = {
            'f_int_sum': {
                'sum': {
                    'field': 'f_nested.f_int'
                }
            },
            'f_int_avg': {
                'avg': {
                    'field': 'f_nested.f_int'
                }
            },
            'f_int_min': {
                'min': {
                    'field': 'f_nested.f_int'
                }
            },
            'f_int_max': {
                'max': {
                    'field': 'f_nested.f_int'
                }
            },
        }
        aggs: 'ESAggsResponse' = await model_nested.aggregate(aggs_def)
        assert aggs['f_int_sum']['value'] == 14.0
        assert aggs['f_int_avg']['value'] == 3.5
        assert aggs['f_int_min']['value'] == 2.0
        assert aggs['f_int_max']['value'] == 5.0

        # Test aggregations with query
        query: 'ESQuery' = {
            'bool': {
                'must': {
                    'match': {
                        'f_nested.f_str': 'nested_test2'
                    }
                }
            }
        }
        aggs: 'ESAggsResponse' = await model_nested.aggregate(aggs_def, query=query)
        assert aggs['f_int_sum']['value'] == 2.0
        assert aggs['f_int_avg']['value'] == 2.0
        assert aggs['f_int_min']['value'] == 2.0
        assert aggs['f_int_max']['value'] == 2.0

        # Test terms aggregation
        aggs_def: 'ESAggs' = {
            'f_int_terms': {
                'terms': {
                    'field': 'f_nested.f_int'
                }
            }
        }
        aggs: 'ESAggsResponse' = await model_nested.aggregate(aggs_def)
        assert aggs['f_int_terms']['buckets'][0]['key'] == 2.0
        assert aggs['f_int_terms']['buckets'][0]['doc_count'] == 1
        assert aggs['f_int_terms']['buckets'][1]['key'] == 3.0
        assert aggs['f_int_terms']['buckets'][1]['doc_count'] == 1
        assert aggs['f_int_terms']['buckets'][2]['key'] == 4.0
        assert aggs['f_int_terms']['buckets'][2]['doc_count'] == 1
        assert aggs['f_int_terms']['buckets'][3]['key'] == 5.0
        assert aggs['f_int_terms']['buckets'][3]['doc_count'] == 1

    async def test_pagination(self, es, esorm, model_nested):
        """
        Test pagination
        """
        pagination = esorm.Pagination(page=1, page_size=2)
        res = await pagination(model_nested).all()
        assert len(res) == 2
        assert res[0].f_nested.f_str == 'nested_test2'
        assert res[1].f_nested.f_str == 'nested_test3'

        pagination = esorm.Pagination(page=2, page_size=2)
        res = await pagination(model_nested).all()
        assert len(res) == 2
        assert res[0].f_nested.f_str == 'nested_test4'
        assert res[1].f_nested.f_str == 'nested_test5'

    async def test_sort(self, es, esorm, model_nested):
        """
        Test sort
        """
        esorm.model.set_max_lazy_property_concurrency(2)

        sort = esorm.Sort(sort=[{'f_nested.f_int': {'order': 'asc'}}])
        res = await sort(model_nested).all()
        assert len(res) == 4
        assert res[0].f_nested.f_int == 2
        assert res[1].f_nested.f_int == 3
        assert res[2].f_nested.f_int == 4
        assert res[3].f_nested.f_int == 5

        sort = esorm.Sort(sort=[{'f_nested.f_int': {'order': 'desc'}}])
        res = await sort(model_nested).all()
        assert len(res) == 4
        assert res[0].f_nested.f_int == 5
        assert res[1].f_nested.f_int == 4
        assert res[2].f_nested.f_int == 3
        assert res[3].f_nested.f_int == 2

        sort = esorm.Sort(sort='f_nested.f_int')
        res = await sort(model_nested).all()
        assert len(res) == 4
        assert res[0].f_nested.f_int == 2
        assert res[1].f_nested.f_int == 3
        assert res[2].f_nested.f_int == 4
        assert res[3].f_nested.f_int == 5

    async def test_lazy_properties(self, es, esorm, model_lazy_prop):
        """
        Test lazy properties
        """
        esorm.model.set_max_lazy_property_concurrency(2)

        # Create 3 documents with the same content
        doc_id1 = await model_lazy_prop(f_str='test').save()
        assert doc_id1 is not None
        doc_id2 = await model_lazy_prop(f_str='test').save()
        assert doc_id2 is not None
        doc_id3 = await model_lazy_prop(f_str='test').save()
        assert doc_id3 is not None
        # Create 2 documents with different content
        doc_id = await model_lazy_prop(f_str='test2').save()
        assert doc_id is not None
        doc_id = await model_lazy_prop(f_str='test3').save()
        assert doc_id is not None

        # Test lazy properties
        sort = esorm.Sort(sort=[{'f_str': {'order': 'asc'}}])
        docs = await sort(model_lazy_prop).all()
        assert len(docs) == 5
        assert docs[0].f_str == 'test'
        same_f_strs = docs[0].same_f_strs
        assert len(same_f_strs) == 3
        assert same_f_strs[0].f_str == 'test'
        assert same_f_strs[0]._id == doc_id1
        assert same_f_strs[1].f_str == 'test'
        assert same_f_strs[1]._id == doc_id2
        assert same_f_strs[2].f_str == 'test'
        assert same_f_strs[2]._id == doc_id3
        assert docs[1].f_str == 'test'
        assert len(docs[1].same_f_strs) == 3
        assert docs[2].f_str == 'test'
        assert len(docs[2].same_f_strs) == 3
        assert docs[3].f_str == 'test2'
        assert len(docs[3].same_f_strs) == 1
        assert docs[4].f_str == 'test3'
        assert len(docs[4].same_f_strs) == 1

    async def test_shard_routing(self, es, esorm, model_config):
        """
        Test shard routing
        """
        doc = model_config(custom_id='test1', f_str='test1')
        doc_id = await doc.save()
        assert doc_id is not None
        query = {
            'bool': {
                'must': {
                    'match': {
                        'f_str': 'test1'
                    }
                }
            }
        }
        doc = await model_config.search_one(query)
        assert doc is not None
        assert doc._id == doc_id == "test1"
        assert doc._routing == "test1_routing"

        doc = model_config(custom_id='test2', f_str='test2')
        doc_id = await doc.save()
        assert doc_id is not None
        query = {
            'bool': {
                'must': {
                    'match': {
                        'f_str': 'test2'
                    }
                }
            }
        }
        docs = await model_config.search(query)
        assert len(docs) == 1
        doc = docs[0]
        assert doc._id == doc_id == "test2"
        assert doc._routing == "test2_routing"

        # Test different routing search
        docs = await model_config.search(query, routing="test1_routing")
        assert len(docs) == 0
        docs = await model_config.search(query, routing="test2_routing")
        assert len(docs) == 1

        # Override routing on save
        doc = model_config(custom_id='test3', f_str='test3')
        doc_id = await doc.save(routing="test3__routing__")
        assert doc_id is not None
        query = {
            'bool': {
                'must': {
                    'match': {
                        'f_str': 'test3'
                    }
                }
            }
        }
        doc = await model_config.search_one(query)
        assert doc is not None
        assert doc._id == doc_id == "test3"
        assert doc._routing == "test3__routing__"

    async def test_binary(self, es, esorm, model_es_optional):
        """
        Test binary fields
        """
        doc = model_es_optional(f_binary=b'\x01\x02\x03\x04')
        doc_id = await doc.save()

        assert doc_id is not None
        doc = await model_es_optional.get(doc_id)
        assert getattr(doc.f_binary, 'bytes') == b'\x01\x02\x03\x04'

        # Modify binary field
        doc.f_binary = b'\x05\x06\x07\x08'
        await doc.save()
        doc = await model_es_optional.get(doc_id)
        assert getattr(doc.f_binary, 'bytes') == b'\x05\x06\x07\x08'

    async def test_binary_nested(self, es, esorm, model_nested_binary):
        """
        Test nested binary fields
        """
        binary_model, nested_binary_model = model_nested_binary
        doc = nested_binary_model(f_nested=binary_model())
        doc.f_nested.f_binary = b'\x01\x02\x03\x04'
        doc_id = await doc.save()
        assert doc_id is not None

    async def test_nested_base_model_lazy_prop(self, es, esorm, model_nested_base_model):
        """
        Test nested base model
        """
        nested_base_model, nested_base_model_model = model_nested_base_model

        assert nested_base_model.ESConfig._lazy_properties['lazy_prop'] is not None

        doc = nested_base_model_model(f_nested=nested_base_model(f_str='test', f_int=1))
        doc_id = await doc.save()
        assert doc_id is not None

        doc = await nested_base_model_model.get(doc_id)
        assert doc is not None
        assert doc.f_nested.f_str == 'test'
        assert doc.f_nested.f_int == 1
        assert doc.f_nested.lazy_prop == 'lazy(f_str=test, f_int=1)'

    async def test_base_model_parent(self, es, esorm, model_base_model_parent):
        """
        Test base model parent
        """
        doc = model_base_model_parent(f_str='test', f_int=123, f_float=0.123)
        doc_id = await doc.save()
        assert doc_id is not None

        doc = await model_base_model_parent.get(doc_id)
        assert doc is not None
        assert doc.f_str == 'test'
        assert doc.f_int == 123
        assert doc.f_float == 0.123

        assert doc._id == 'test'



import requests


def generate_changelog(repository, file_path="changelog.rst"):
    """
    Generate a changelog in reStructuredText format from GitHub releases of a public repository.
    The changelog will have a single main title 'Changelog' and each release as a subsection.

    :param repository: The GitHub repository in the format 'username/repo'.
    :param file_path: The path of the changelog file.
    """
    # GitHub API URL for the releases of the repository
    api_url = f"https://api.github.com/repos/{repository}/releases"

    # Send a request to the GitHub API
    response = requests.get(api_url)

    # Check if the response is successful
    if response.status_code == 200:
        releases = response.json()

        # Main title for the changelog
        changelog = ["Changelog\n=========\n\n"]

        # Process each release
        for release in releases:
            title = release["name"]
            date = release["published_at"].split("T")[0]
            body = release["body"].replace('\r\n', '\n').strip()

            # Format each release as a subsection
            release_entry = f"{title}\n{'-' * len(title)}\n*Released on*: {date}\n\n{body}\n\n"
            changelog.append(release_entry)

        # Join all entries into a single string
        changelog = "\n".join(changelog)

        # Save changelog to file
        with open(file_path, 'w') as file:
            file.write(changelog)
    else:
        response.raise_for_status()
